# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2011 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD Panel"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package ArchPanel
#  \ingroup ARCH
#  \brief The Panel object and tools
#
#  This module provides tools to build Panel objects.
#  Panels consist of a closed shape that gets extruded to
#  produce a flat object.

import math

import FreeCAD
import ArchCommands
import ArchComponent
import Draft
import DraftVecUtils
import Part

from FreeCAD import Vector
from draftutils import params

if FreeCAD.GuiUp:
    from PySide import QtCore, QtGui
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import FreeCADGui
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt

    # \endcond


class _Panel(ArchComponent.Component):
    "The Panel object"

    def __init__(self, obj):

        ArchComponent.Component.__init__(self, obj)
        self.Type = "Panel"
        self.setProperties(obj)
        obj.IfcType = "Plate"

    def setProperties(self, obj):

        pl = obj.PropertiesList
        if not "Length" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Length",
                "Panel",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The length of this element, if not based on a profile"
                ),
                locked=True,
            )
        if not "Width" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Width",
                "Panel",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The width of this element, if not based on a profile"
                ),
                locked=True,
            )
        if not "Thickness" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Thickness",
                "Panel",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The thickness or extrusion depth of this element"
                ),
                locked=True,
            )
        if not "Sheets" in pl:
            obj.addProperty(
                "App::PropertyInteger",
                "Sheets",
                "Panel",
                QT_TRANSLATE_NOOP("App::Property", "The number of sheets to use"),
                locked=True,
            )
            obj.Sheets = 1
        if not "Offset" in pl:
            obj.addProperty(
                "App::PropertyDistance",
                "Offset",
                "Panel",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The offset between this panel and its baseline"
                ),
                locked=True,
            )
        if not "WaveLength" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "WaveLength",
                "Panel",
                QT_TRANSLATE_NOOP("App::Property", "The length of waves for corrugated elements"),
                locked=True,
            )
        if not "WaveHeight" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "WaveHeight",
                "Panel",
                QT_TRANSLATE_NOOP("App::Property", "The height of waves for corrugated elements"),
                locked=True,
            )
        if not "WaveOffset" in pl:
            obj.addProperty(
                "App::PropertyDistance",
                "WaveOffset",
                "Panel",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The horizontal offset of waves for corrugated elements"
                ),
                locked=True,
            )
        if not "WaveDirection" in pl:
            obj.addProperty(
                "App::PropertyAngle",
                "WaveDirection",
                "Panel",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The direction of waves for corrugated elements"
                ),
                locked=True,
            )
        if not "WaveType" in pl:
            obj.addProperty(
                "App::PropertyEnumeration",
                "WaveType",
                "Panel",
                QT_TRANSLATE_NOOP("App::Property", "The type of waves for corrugated elements"),
                locked=True,
            )
            obj.WaveType = ["Curved", "Trapezoidal", "Spikes"]
        if not "WaveBottom" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "WaveBottom",
                "Panel",
                QT_TRANSLATE_NOOP(
                    "App::Property", "If the wave also affects the bottom side or not"
                ),
                locked=True,
            )
        if not "Area" in pl:
            obj.addProperty(
                "App::PropertyArea",
                "Area",
                "Panel",
                QT_TRANSLATE_NOOP("App::Property", "The area of this panel"),
                locked=True,
            )
        if not "FaceMaker" in pl:
            obj.addProperty(
                "App::PropertyEnumeration",
                "FaceMaker",
                "Panel",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The facemaker type to use to build the profile of this object"
                ),
                locked=True,
            )
            obj.FaceMaker = ["None", "Simple", "Cheese", "Bullseye"]
        if not "Normal" in pl:
            obj.addProperty(
                "App::PropertyVector",
                "Normal",
                "Panel",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The normal extrusion direction of this object (keep (0,0,0) for automatic normal)",
                ),
                locked=True,
            )
        obj.setEditorMode("VerticalArea", 2)
        obj.setEditorMode("HorizontalArea", 2)

    def onDocumentRestored(self, obj):

        ArchComponent.Component.onDocumentRestored(self, obj)
        self.setProperties(obj)

    def loads(self, state):

        self.Type = "Panel"

    def execute(self, obj):
        "creates the panel shape"

        if self.clone(obj):
            return

        layers = []
        length = 0
        width = 0
        thickness = 0

        # base tests
        if obj.Base:
            if hasattr(obj.Base, "Shape"):
                if obj.Base.Shape.isNull():
                    return
            elif obj.Base.isDerivedFrom("Mesh::Feature"):
                if not obj.Base.Mesh.isSolid():
                    return
        else:
            if obj.Length.Value:
                length = obj.Length.Value
            else:
                return
            if obj.Width.Value:
                width = obj.Width.Value
            else:
                return
        if obj.Thickness.Value:
            thickness = obj.Thickness.Value
        else:
            if not obj.Base:
                return
            elif hasattr(obj.Base, "Shape"):
                if not obj.Base.Shape.Solids:
                    return
        if hasattr(obj, "Material"):
            if obj.Material:
                if hasattr(obj.Material, "Materials"):
                    varwidth = 0
                    thicknesses = [t for t in obj.Material.Thicknesses if t >= 0]
                    restwidth = thickness - sum(thicknesses)
                    if restwidth > 0:
                        varwidth = [t for t in thicknesses if t == 0]
                        if varwidth:
                            varwidth = restwidth / len(varwidth)
                    for t in obj.Material.Thicknesses:
                        if t:
                            layers.append(t)
                        elif varwidth:
                            layers.append(varwidth)
        # creating base shape
        pl = obj.Placement
        base = None
        normal = None
        if hasattr(obj, "Normal"):
            if obj.Normal.Length > 0:
                normal = Vector(obj.Normal)
                normal.normalize()
                normal.multiply(thickness)
        baseprofile = None
        if obj.Base:
            base = obj.Base.Shape.copy()
            if not base.Solids:
                # p = FreeCAD.Placement(obj.Base.Placement)
                if base.Faces:
                    baseprofile = base
                    if not normal:
                        normal = baseprofile.Faces[0].normalAt(0, 0).multiply(thickness)
                    if layers:
                        layeroffset = 0
                        shps = []
                        for l in layers:
                            if l >= 0:
                                n = Vector(normal).normalize().multiply(abs(l))
                                b = base.extrude(n)
                                if layeroffset:
                                    o = Vector(normal).normalize().multiply(layeroffset)
                                    b.translate(o)
                                shps.append(b)
                            layeroffset += abs(l)
                        base = Part.makeCompound(shps)
                    else:
                        base = base.extrude(normal)
                elif base.Wires:
                    fm = False
                    if hasattr(obj, "FaceMaker"):
                        if obj.FaceMaker != "None":
                            try:
                                baseprofile = Part.makeFace(
                                    base.Wires, "Part::FaceMaker" + str(obj.FaceMaker)
                                )
                                fm = True
                            except Exception:
                                FreeCAD.Console.PrintError(
                                    translate("Arch", "Facemaker returned an error") + "\n"
                                )
                                return
                    if not fm:
                        closed = True
                        for w in base.Wires:
                            if not w.isClosed():
                                closed = False
                        if closed:
                            baseprofile = ArchCommands.makeFace(base.Wires)
                    if not normal:
                        normal = baseprofile.normalAt(0, 0).multiply(thickness)
                    if layers:
                        layeroffset = 0
                        shps = []
                        for l in layers:
                            if l >= 0:
                                n = Vector(normal).normalize().multiply(abs(l))
                                b = baseprofile.extrude(n)
                                if layeroffset:
                                    o = Vector(normal).normalize().multiply(layeroffset)
                                    b.translate(o)
                                shps.append(b)
                            layeroffset += abs(l)
                        base = Part.makeCompound(shps)
                    else:
                        base = baseprofile.extrude(normal)
                elif obj.Base.isDerivedFrom("Mesh::Feature"):
                    if obj.Base.Mesh.isSolid():
                        if obj.Base.Mesh.countComponents() == 1:
                            sh = ArchCommands.getShapeFromMesh(obj.Base.Mesh)
                            if sh.isClosed() and sh.isValid() and sh.Solids:
                                base = sh
        else:
            if layers:
                shps = []
                layeroffset = 0
                for l in layers:
                    if l >= 0:
                        if normal:
                            n = Vector(normal).normalize().multiply(l)
                        else:
                            n = Vector(0, 0, 1).multiply(abs(l))
                        l2 = length / 2 or 0.5
                        w2 = width / 2 or 0.5
                        v1 = Vector(-l2, -w2, layeroffset)
                        v2 = Vector(l2, -w2, layeroffset)
                        v3 = Vector(l2, w2, layeroffset)
                        v4 = Vector(-l2, w2, layeroffset)
                        base = Part.makePolygon([v1, v2, v3, v4, v1])
                        baseprofile = Part.Face(base)
                        base = baseprofile.extrude(n)
                        shps.append(base)
                    layeroffset += abs(l)
                base = Part.makeCompound(shps)
            else:
                if not normal:
                    normal = Vector(0, 0, 1).multiply(thickness)
                l2 = length / 2 or 0.5
                w2 = width / 2 or 0.5
                v1 = Vector(-l2, -w2, 0)
                v2 = Vector(l2, -w2, 0)
                v3 = Vector(l2, w2, 0)
                v4 = Vector(-l2, w2, 0)
                base = Part.makePolygon([v1, v2, v3, v4, v1])
                baseprofile = Part.Face(base)
                base = baseprofile.extrude(normal)

        if hasattr(obj, "Area"):
            if baseprofile:
                obj.Area = baseprofile.Area

        if hasattr(obj, "WaveLength"):
            if baseprofile and obj.WaveLength.Value and obj.WaveHeight.Value:
                # corrugated element
                bb = baseprofile.BoundBox
                bb.enlarge(bb.DiagonalLength)
                downsegment = None
                if hasattr(obj, "WaveBottom"):
                    if not obj.WaveBottom:
                        if obj.WaveType == "Curved":
                            if obj.Thickness.Value > obj.WaveHeight.Value:
                                downsegment = obj.Thickness.Value
                            else:
                                downsegment = obj.WaveHeight.Value + obj.Thickness.Value
                        else:
                            downsegment = obj.Thickness.Value
                p1 = Vector(0, 0, 0)
                p5 = Vector(obj.WaveLength.Value * 2, 0, 0)
                if obj.WaveType == "Curved":
                    p2 = Vector(obj.WaveLength.Value / 2, 0, obj.WaveHeight.Value)
                    p3 = Vector(obj.WaveLength.Value, 0, 0)
                    e1 = Part.Arc(p1, p2, p3).toShape()
                    p4 = Vector(obj.WaveLength.Value * 1.5, 0, -obj.WaveHeight.Value)
                    e2 = Part.Arc(p3, p4, p5).toShape()
                    upsegment = Part.Wire([e1, e2])
                    if not downsegment:
                        if obj.Thickness.Value < e1.Curve.Radius:
                            c3 = e1.Curve.copy()
                            c3.Radius = e1.Curve.Radius - obj.Thickness.Value
                            e3 = Part.Arc(c3, e1.FirstParameter, e1.LastParameter).toShape()
                            c4 = e2.Curve.copy()
                            c4.Radius = e2.Curve.Radius + obj.Thickness.Value
                            e4 = Part.Arc(c4, e2.FirstParameter, e2.LastParameter).toShape()
                            downsegment = Part.Wire([e3, e4])
                        else:
                            r = e2.Curve.Radius + obj.Thickness.Value
                            z = math.sqrt(r**2 - obj.WaveLength.Value**2)
                            p6 = e2.Curve.Center.add(Vector(-obj.WaveLength, 0, -z))
                            p7 = e2.Curve.Center.add(Vector(0, 0, -r))
                            p8 = e2.Curve.Center.add(Vector(obj.WaveLength, 0, -z))
                            downsegment = Part.Arc(p6, p7, p8).toShape()

                elif obj.WaveType == "Trapezoidal":
                    p2 = Vector(obj.WaveLength.Value / 4, 0, obj.WaveHeight.Value)
                    p3 = Vector(obj.WaveLength.Value, 0, obj.WaveHeight.Value)
                    p4 = Vector(obj.WaveLength.Value * 1.25, 0, 0)
                    upsegment = Part.makePolygon([p1, p2, p3, p4, p5])
                    if not downsegment:
                        a = ((p1.sub(p2)).getAngle(p3.sub(p2))) / 2
                        tx = obj.Thickness.Value / math.tan(a)
                        d1 = Vector(tx, 0, -obj.Thickness.Value)
                        d2 = Vector(-tx, 0, -obj.Thickness.Value)
                        p6 = p1.add(d1)
                        if tx >= p3.sub(p2).Length / 2:
                            d3 = p2.sub(p1)
                            d3.normalize()
                            d3.multiply((0.625 * obj.WaveLength.Value) / d3.x)
                            d4 = Vector(d3.x, 0, -d3.z)
                            p7 = p6.add(d3)
                            p8 = p7.add(d4)
                            p9 = p5.add(d1)
                            downsegment = Part.makePolygon([p6, p7, p8, p9])
                        elif tx <= 0.625 * obj.WaveLength.Value:
                            p7 = p2.add(d1)
                            p8 = p3.add(d2)
                            p9 = p4.add(d2)
                            p10 = p5.add(d1)
                            downsegment = Part.makePolygon([p6, p7, p8, p9, p10])
                        else:
                            downsegment = obj.Thickness.Value

                else:  # spike
                    p2 = Vector(obj.WaveHeight.Value, 0, obj.WaveHeight.Value)
                    p3 = Vector(obj.WaveHeight.Value * 2, 0, 0)
                    upsegment = Part.makePolygon([p1, p2, p3, p5])
                    if not downsegment:
                        downsegment = obj.Thickness.Value

                upsegment.translate(Vector(bb.getPoint(0).x, bb.getPoint(0).y, bb.Center.z))
                if isinstance(downsegment, Part.Shape):
                    downsegment.translate(Vector(bb.getPoint(0).x, bb.getPoint(0).y, bb.Center.z))
                if hasattr(obj, "WaveOffset"):
                    if obj.WaveOffset.Value:
                        upsegment.translate(Vector(obj.WaveOffset.Value, 0, 0))
                        if isinstance(downsegment, Part.Shape):
                            downsegment.translate(Vector(obj.WaveOffset.Value, 0, 0))

                upedges = []
                downedges = []
                for i in range(int(bb.XLength / (obj.WaveLength.Value * 2))):
                    w1 = upsegment.copy()
                    w1.translate(Vector(obj.WaveLength.Value * 2 * i, 0, 0))
                    upedges.extend(w1.Edges)
                    if isinstance(downsegment, Part.Shape):
                        w2 = downsegment.copy()
                        w2.translate(Vector(obj.WaveLength.Value * 2 * i, 0, 0))
                        downedges.extend(w2.Edges)
                upwire = Part.Wire(upedges)
                FreeCAD.upwire = upwire  # REMOVE
                if isinstance(downsegment, Part.Shape):
                    downwire = Part.Wire(downedges)
                    FreeCAD.downwire = downwire  # REMOVE
                    e1 = Part.LineSegment(
                        upwire.Vertexes[0].Point, downwire.Vertexes[0].Point
                    ).toShape()
                    e2 = Part.LineSegment(
                        upwire.Vertexes[-1].Point, downwire.Vertexes[-1].Point
                    ).toShape()
                    basewire = Part.Wire(upwire.Edges + [e1, e2] + downwire.Edges)
                else:
                    z = obj.Thickness.Value
                    if obj.WaveType == "Curved":
                        z += obj.WaveHeight.Value
                    p1 = upwire.Vertexes[0].Point
                    p2 = p1.add(Vector(0, 0, -z))
                    p3 = Vector(upwire.Vertexes[-1].Point.x, upwire.Vertexes[-1].Point.y, p2.z)
                    p4 = upwire.Vertexes[-1].Point
                    w = Part.makePolygon([p1, p2, p3, p4])
                    basewire = Part.Wire(upwire.Edges + w.Edges)

                FreeCAD.basewire = basewire
                if not basewire.isClosed():
                    print("Error closing base wire - check FreeCAD.basewire")
                    return

                baseface = Part.Face(basewire)
                base = baseface.extrude(Vector(0, bb.YLength, 0))
                if normal.cross(FreeCAD.Vector(0, 0, 1)).Length > 1e-6:
                    rot = FreeCAD.Rotation(FreeCAD.Vector(0, -1, 0), FreeCAD.Vector(*normal[:2], 0))
                    base.rotate(bb.Center, rot.Axis, math.degrees(rot.Angle))
                rot = FreeCAD.Rotation(FreeCAD.Vector(0, 0, 1), normal)
                base.rotate(bb.Center, rot.Axis, math.degrees(rot.Angle))
                if obj.WaveDirection.Value:
                    base.rotate(bb.Center, normal, obj.WaveDirection.Value)
                n1 = normal.negative().normalize().multiply(obj.WaveHeight.Value * 2)
                self.vol = baseprofile.copy()
                self.vol.translate(n1)
                self.vol = self.vol.extrude(n1.negative().multiply(2))
                base = self.vol.common(base)
                base = base.removeSplitter()
                if not base:
                    FreeCAD.Console.PrintError(
                        translate("Arch", "Error computing shape of") + " " + obj.Label + "\n"
                    )
                    return False

        if base and (obj.Sheets > 1) and normal and thickness:
            bases = [base]
            for i in range(1, obj.Sheets):
                n = FreeCAD.Vector(normal).normalize().multiply(i * thickness)
                b = base.copy()
                b.translate(n)
                bases.append(b)
            base = Part.makeCompound(bases)

        if base and normal and hasattr(obj, "Offset"):
            if obj.Offset.Value:
                v = DraftVecUtils.scaleTo(normal, obj.Offset.Value)
                base.translate(v)

        # process subshapes
        base = self.processSubShapes(obj, base, pl)

        # applying
        if base:
            if not base.isNull():
                if base.isValid() and base.Solids:
                    if len(base.Solids) == 1:
                        if base.Volume < 0:
                            base.reverse()
                        if base.Volume < 0:
                            FreeCAD.Console.PrintError(
                                translate("Arch", "Could not compute a shape")
                            )
                            return
                        base = base.removeSplitter()
                    obj.Shape = base
                    if not pl.isNull():
                        obj.Placement = pl


class _ViewProviderPanel(ArchComponent.ViewProviderComponent):
    "A View Provider for the Panel object"

    def __init__(self, vobj):

        ArchComponent.ViewProviderComponent.__init__(self, vobj)
        vobj.ShapeColor = ArchCommands.getDefaultColor("Panel")

    def getIcon(self):

        # import Arch_rc
        if hasattr(self, "Object"):
            if hasattr(self.Object, "CloneOf"):
                if self.Object.CloneOf:
                    return ":/icons/Arch_Panel_Clone.svg"
        return ":/icons/Arch_Panel_Tree.svg"

    def updateData(self, obj, prop):

        if prop in ["Placement", "Shape", "Material"]:
            if hasattr(obj, "Material"):
                if obj.Material:
                    if hasattr(obj.Material, "Materials"):
                        activematerials = [
                            obj.Material.Materials[i]
                            for i in range(len(obj.Material.Materials))
                            if obj.Material.Thicknesses[i] >= 0
                        ]
                        if len(activematerials) == len(obj.Shape.Solids):
                            cols = []
                            for i, mat in enumerate(activematerials):
                                c = obj.ViewObject.ShapeColor
                                c = (c[0], c[1], c[2], 1.0 - obj.ViewObject.Transparency / 100.0)
                                if "DiffuseColor" in mat.Material:
                                    if "(" in mat.Material["DiffuseColor"]:
                                        c = tuple(
                                            [
                                                float(f)
                                                for f in mat.Material["DiffuseColor"]
                                                .strip("()")
                                                .split(",")
                                            ]
                                        )
                                if "Transparency" in mat.Material:
                                    c = (
                                        c[0],
                                        c[1],
                                        c[2],
                                        1.0 - float(mat.Material["Transparency"]),
                                    )
                                cols.extend([c for j in range(len(obj.Shape.Solids[i].Faces))])
                            if obj.ViewObject.DiffuseColor != cols:
                                obj.ViewObject.DiffuseColor = cols
        ArchComponent.ViewProviderComponent.updateData(self, obj, prop)


class PanelCut(Draft.DraftObject):
    "A flat, 2D view of an Arch Panel"

    def __init__(self, obj):
        Draft.DraftObject.__init__(self, obj)
        obj.Proxy = self

        # setProperties of ArchComponent will be overwritten
        # thus setProperties from ArchComponent will be explicit called to get the properties
        ArchComponent.ViewProviderComponent.setProperties(self, obj)

        self.setProperties(obj)

    def setProperties(self, obj):

        pl = obj.PropertiesList
        if not "Source" in pl:
            obj.addProperty(
                "App::PropertyLink",
                "Source",
                "PanelCut",
                QT_TRANSLATE_NOOP("App::Property", "The linked object"),
                locked=True,
            )
        if not "TagText" in pl:
            obj.addProperty(
                "App::PropertyString",
                "TagText",
                "PanelCut",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The text to display. Can be %tag%, %label% or %description% to display the panel tag or label",
                ),
                locked=True,
            )
            obj.TagText = "%tag%"
        if not "TagSize" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "TagSize",
                "PanelCut",
                QT_TRANSLATE_NOOP("App::Property", "The size of the tag text"),
                locked=True,
            )
            obj.TagSize = 10
        if not "TagPosition" in pl:
            obj.addProperty(
                "App::PropertyVector",
                "TagPosition",
                "PanelCut",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The position of the tag text. Keep (0,0,0) for center position",
                ),
                locked=True,
            )
        if not "TagRotation" in pl:
            obj.addProperty(
                "App::PropertyAngle",
                "TagRotation",
                "PanelCut",
                QT_TRANSLATE_NOOP("App::Property", "The rotation of the tag text"),
                locked=True,
            )
        if not "FontFile" in pl:
            obj.addProperty(
                "App::PropertyFile",
                "FontFile",
                "PanelCut",
                QT_TRANSLATE_NOOP("App::Property", "The font of the tag text"),
                locked=True,
            )
            obj.FontFile = params.get_param("ShapeStringFontFile")
        if not "MakeFace" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "MakeFace",
                "PanelCut",
                QT_TRANSLATE_NOOP(
                    "App::Property", "If True, the object is rendered as a face, if possible."
                ),
                locked=True,
            )
        if not "AllowedAngles" in pl:
            obj.addProperty(
                "App::PropertyFloatList",
                "AllowedAngles",
                "PanelCut",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The allowed angles this object can be rotated to when placed on sheets",
                ),
                locked=True,
            )
        self.Type = "PanelCut"
        if not "CutOffset" in pl:
            obj.addProperty(
                "App::PropertyDistance",
                "CutOffset",
                "PanelCut",
                QT_TRANSLATE_NOOP(
                    "App::Property", "An offset value to move the cut plane from the center point"
                ),
                locked=True,
            )

    def onDocumentRestored(self, obj):

        self.setProperties(obj)

    def execute(self, obj):

        pl = obj.Placement
        if obj.Source:
            base = None
            n = None
            if Draft.getType(obj.Source) == "Panel":
                import DraftGeomUtils
                import Part

                baseobj = None
                if obj.Source.CloneOf:
                    baseobj = obj.Source.CloneOf.Base
                if obj.Source.Base:
                    baseobj = obj.Source.Base
                if baseobj:
                    if hasattr(baseobj, "Shape"):
                        if baseobj.Shape.Solids:
                            center = baseobj.Shape.BoundBox.Center
                            diag = baseobj.Shape.BoundBox.DiagonalLength
                            if obj.Source.Normal.Length:
                                n = obj.Source.Normal
                            elif baseobj.isDerivedFrom("Part::Extrusion"):
                                n = baseobj.Dir
                            if not n:
                                n = Vector(0, 0, 1)
                            if hasattr(obj, "CutOffset") and obj.CutOffset.Value:
                                l = obj.CutOffset.Value
                                d = Vector(n)
                                d.multiply(l)
                                center = center.add(d)
                            plane = Part.makePlane(diag, diag, center, n)
                            plane.translate(center.sub(plane.BoundBox.Center))
                            wires = []
                            for sol in baseobj.Shape.Solids:
                                s = sol.section(plane)
                                wires.extend(DraftGeomUtils.findWires(s.Edges))
                            if wires:
                                base = self.buildCut(obj, wires)
                        else:
                            base = self.buildCut(obj, baseobj.Shape.Wires)
                            for w in base.Wires:
                                n = DraftGeomUtils.getNormal(w)
                                if n:
                                    break
                            if not n:
                                n = Vector(0, 0, 1)
                        if base and n:
                            base.translate(base.BoundBox.Center.negative())
                            r = FreeCAD.Rotation(n, Vector(0, 0, 1))
                            base.rotate(Vector(0, 0, 0), r.Axis, math.degrees(r.Angle))
                    elif baseobj.isDerivedFrom("Mesh::Feature"):
                        return
                else:
                    l2 = obj.Source.Length / 2
                    w2 = obj.Source.Width / 2
                    v1 = Vector(-l2, -w2, 0)
                    v2 = Vector(l2, -w2, 0)
                    v3 = Vector(l2, w2, 0)
                    v4 = Vector(-l2, w2, 0)
                    base = Part.makePolygon([v1, v2, v3, v4, v1])
                if base:
                    self.outline = base
                    if obj.FontFile and obj.TagText and obj.TagSize.Value:
                        if obj.TagPosition.Length == 0:
                            pos = base.BoundBox.Center
                        else:
                            pos = obj.TagPosition
                        if obj.TagText == "%tag%":
                            string = obj.Source.Tag
                        elif obj.TagText == "%label%":
                            string = obj.Source.Label
                        elif obj.TagText == "%description%":
                            string = obj.Source.Description
                        else:
                            string = obj.TagText
                        chars = []
                        for char in Part.makeWireString(string, obj.FontFile, obj.TagSize.Value, 0):
                            chars.extend(char)
                        textshape = Part.Compound(chars)
                        textshape.translate(pos.sub(textshape.BoundBox.Center))
                        textshape.rotate(
                            textshape.BoundBox.Center, Vector(0, 0, 1), obj.TagRotation.Value
                        )
                        self.tag = textshape
                        base = Part.Compound([base, textshape])
                    else:
                        base = Part.Compound([base])
                    obj.Shape = base
                    obj.Placement = pl

    def buildCut(self, obj, wires):
        """buildCut(obj,wires): builds the object shape"""

        import Part

        if hasattr(obj, "MakeFace"):
            if obj.MakeFace:
                face = None
                if len(wires) > 1:
                    d = 0
                    ow = None
                    for w in wires:
                        if w.BoundBox.DiagonalLength > d:
                            d = w.BoundBox.DiagonalLength
                            ow = w
                    if ow:
                        face = Part.Face(ow)
                        for w in wires:
                            if w.hashCode() != ow.hashCode():
                                wface = Part.Face(w)
                                face = face.cut(wface)
                else:
                    face = Part.Face(wires[0])
                if face:
                    return face
        return Part.makeCompound(wires)

    def getWires(self, obj):
        """getWires(obj): returns a tuple containing 3 shapes
        that define the panel outline, the panel holes, and
        tags (engravings): (outline,holes,tags). Any of these can
        be None if nonexistent"""

        tag = None
        outl = None
        inl = None
        if not hasattr(self, "outline"):
            self.execute(obj)
        if not hasattr(self, "outline"):
            return None
        outl = self.outline.copy()
        if hasattr(self, "tag"):
            tag = self.tag.copy()
        if tag:
            tag.Placement = obj.Placement.multiply(tag.Placement)

        outl = self.outline.copy()
        outl.Placement = obj.Placement.multiply(outl.Placement)
        if len(outl.Wires) > 1:
            # separate outline
            d = 0
            ow = None
            for w in outl.Wires:
                if w.BoundBox.DiagonalLength > d:
                    d = w.BoundBox.DiagonalLength
                    ow = w
            if ow:
                inl = Part.Compound([w for w in outl.Wires if w.hashCode() != ow.hashCode()])
                outl = Part.Compound([ow])
        else:
            inl = None
            outl = Part.Compound([outl.Wires[0]])
        return (outl, inl, tag)


class ViewProviderPanelCut(Draft.ViewProviderDraft):
    "a view provider for the panel cut object"

    def __init__(self, vobj):

        Draft.ViewProviderDraft.__init__(self, vobj)
        self.setProperties(vobj)

    def setProperties(self, vobj):

        pl = vobj.PropertiesList
        if not "Margin" in pl:
            vobj.addProperty(
                "App::PropertyLength",
                "Margin",
                "Arch",
                QT_TRANSLATE_NOOP("App::Property", "A margin inside the boundary"),
                locked=True,
            )
        if not "ShowMargin" in pl:
            vobj.addProperty(
                "App::PropertyBool",
                "ShowMargin",
                "Arch",
                QT_TRANSLATE_NOOP("App::Property", "Turns the display of the margin on/off"),
                locked=True,
            )

    def onDocumentRestored(self, vobj):

        self.setProperties(vobj)

    def attach(self, vobj):

        Draft.ViewProviderDraft.attach(self, vobj)
        from pivy import coin

        self.coords = coin.SoCoordinate3()
        self.lineset = coin.SoLineSet()
        self.lineset.numVertices.setValue(-1)
        lineStyle = coin.SoDrawStyle()
        lineStyle.linePattern = 0x0F0F
        self.color = coin.SoBaseColor()
        self.switch = coin.SoSwitch()
        sep = coin.SoSeparator()
        self.switch.whichChild = -1
        sep.addChild(self.color)
        sep.addChild(lineStyle)
        sep.addChild(self.coords)
        sep.addChild(self.lineset)
        self.switch.addChild(sep)
        vobj.Annotation.addChild(self.switch)
        self.onChanged(vobj, "ShowMargin")
        self.onChanged(vobj, "LineColor")

    def onChanged(self, vobj, prop):

        if prop in ["Margin", "ShowMargin"]:
            if hasattr(vobj, "Margin") and hasattr(vobj, "ShowMargin"):
                if (vobj.Margin.Value > 0) and vobj.Object.Shape and vobj.ShowMargin:
                    self.lineset.numVertices.setValue(-1)
                    if vobj.Object.Shape.Wires:
                        d = 0
                        dw = None
                        for w in vobj.Object.Shape.Wires:
                            if w.BoundBox.DiagonalLength > d:
                                d = w.BoundBox.DiagonalLength
                                dw = w
                        if dw:
                            ow = dw.makeOffset2D(vobj.Margin.Value)
                            verts = []
                            for v in ow.OrderedVertexes:
                                v = vobj.Object.Placement.inverse().multVec(v.Point)
                                verts.append((v.x, v.y, v.z))
                            if dw.isClosed():
                                verts.append(verts[0])
                        self.coords.point.setValues(verts)
                        self.lineset.numVertices.setValue(len(verts))
                        self.switch.whichChild = 0
                else:
                    self.switch.whichChild = -1
        elif prop == "LineColor":
            if hasattr(vobj, "LineColor"):
                c = vobj.LineColor
                self.color.rgb.setValue(c[0], c[1], c[2])
        Draft.ViewProviderDraft.onChanged(self, vobj, prop)

    def updateData(self, obj, prop):

        if prop in ["Shape"]:
            self.onChanged(obj.ViewObject, "Margin")
        Draft.ViewProviderDraft.updateData(self, obj, prop)

    def doubleClicked(self, vobj):

        # See setEdit in ViewProviderDraft.
        FreeCADGui.runCommand("Std_TransformManip")
        return True


class PanelSheet(Draft.DraftObject):
    "A collection of Panel cuts under a sheet"

    def __init__(self, obj):

        Draft.DraftObject.__init__(self, obj)
        obj.Proxy = self
        self.setProperties(obj)

    def setProperties(self, obj):

        pl = obj.PropertiesList
        if not "Group" in pl:
            obj.addProperty(
                "App::PropertyLinkList",
                "Group",
                "PanelSheet",
                QT_TRANSLATE_NOOP("App::Property", "The linked Panel cuts"),
                locked=True,
            )
        if not "TagText" in pl:
            obj.addProperty(
                "App::PropertyString",
                "TagText",
                "PanelSheet",
                QT_TRANSLATE_NOOP("App::Property", "The tag text to display"),
                locked=True,
            )
        if not "TagSize" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "TagSize",
                "PanelSheet",
                QT_TRANSLATE_NOOP("App::Property", "The size of the tag text"),
                locked=True,
            )
            obj.TagSize = 10
        if not "TagPosition" in pl:
            obj.addProperty(
                "App::PropertyVector",
                "TagPosition",
                "PanelSheet",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The position of the tag text. Keep (0,0,0) for center position",
                ),
                locked=True,
            )
        if not "TagRotation" in pl:
            obj.addProperty(
                "App::PropertyAngle",
                "TagRotation",
                "PanelSheet",
                QT_TRANSLATE_NOOP("App::Property", "The rotation of the tag text"),
                locked=True,
            )
        if not "FontFile" in pl:
            obj.addProperty(
                "App::PropertyFile",
                "FontFile",
                "PanelSheet",
                QT_TRANSLATE_NOOP("App::Property", "The font of the tag text"),
                locked=True,
            )
            obj.FontFile = params.get_param("ShapeStringFontFile")
        if not "Width" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Width",
                "PanelSheet",
                QT_TRANSLATE_NOOP("App::Property", "The width of the sheet"),
                locked=True,
            )
            obj.Width = params.get_param_arch("PanelLength")
        if not "Height" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Height",
                "PanelSheet",
                QT_TRANSLATE_NOOP("App::Property", "The height of the sheet"),
                locked=True,
            )
            obj.Height = params.get_param_arch("PanelWidth")
        if not "FillRatio" in pl:
            obj.addProperty(
                "App::PropertyPercent",
                "FillRatio",
                "PanelSheet",
                QT_TRANSLATE_NOOP("App::Property", "The fill ratio of this sheet"),
                locked=True,
            )
            obj.setEditorMode("FillRatio", 2)
        if not "MakeFace" in pl:
            obj.addProperty(
                "App::PropertyBool",
                "MakeFace",
                "PanelSheet",
                QT_TRANSLATE_NOOP(
                    "App::Property", "If True, the object is rendered as a face, if possible."
                ),
                locked=True,
            )
        if not "GrainDirection" in pl:
            obj.addProperty(
                "App::PropertyAngle",
                "GrainDirection",
                "PanelSheet",
                QT_TRANSLATE_NOOP(
                    "App::Property", "Specifies an angle for the wood grain (Clockwise, 0 is North)"
                ),
                locked=True,
            )
        if not "Scale" in pl:
            obj.addProperty(
                "App::PropertyFloat",
                "Scale",
                "PanelSheet",
                QT_TRANSLATE_NOOP(
                    "App::Property", "Specifies the scale applied to each panel view."
                ),
                locked=True,
            )
            obj.Scale = 1.0
        if not "Rotations" in pl:
            obj.addProperty(
                "App::PropertyFloatList",
                "Rotations",
                "PanelSheet",
                QT_TRANSLATE_NOOP("App::Property", "A list of possible rotations for the nester"),
                locked=True,
            )
        self.Type = "PanelSheet"

    def onDocumentRestored(self, obj):

        self.setProperties(obj)

    def execute(self, obj):

        import Part

        self.sheettag = None
        self.sheetborder = None
        pl = obj.Placement
        if obj.Width.Value and obj.Height.Value:
            l2 = obj.Width.Value / 2
            w2 = obj.Height.Value / 2
            v1 = Vector(-l2, -w2, 0)
            v2 = Vector(l2, -w2, 0)
            v3 = Vector(l2, w2, 0)
            v4 = Vector(-l2, w2, 0)
            base = Part.makePolygon([v1, v2, v3, v4, v1])
            if hasattr(obj, "MakeFace"):
                if obj.MakeFace:
                    base = Part.Face(base)
            self.sheetborder = base
            wires = []
            area = obj.Width.Value * obj.Height.Value
            subarea = 0
            for v in obj.Group:
                if hasattr(v, "Shape"):
                    wires.extend(v.Shape.Wires)
                    if Draft.getType(v) == "PanelCut":
                        if v.Source:
                            subarea += v.Source.Area.Value
                    else:
                        for w in v.Shape.Wires:
                            if w.isClosed():
                                f = Part.Face(w)
                                subarea += f.Area
            if wires:
                base = Part.Compound([base] + wires)
            if obj.FontFile and obj.TagText and obj.TagSize.Value:
                chars = []
                for char in Part.makeWireString(obj.TagText, obj.FontFile, obj.TagSize.Value, 0):
                    chars.extend(char)
                textshape = Part.Compound(chars)
                textshape.translate(obj.TagPosition)
                textshape.rotate(textshape.BoundBox.Center, Vector(0, 0, 1), obj.TagRotation.Value)
                self.sheettag = textshape
                base = Part.Compound([base, textshape])
            base.scale(obj.Scale, FreeCAD.Vector())
            obj.Shape = base
            obj.Placement = pl
            obj.FillRatio = int((subarea / area) * 100)

    def getOutlines(self, obj, transform=False):
        """getOutlines(obj,transform=False): returns a list of compounds whose wires define the
        outlines of the panels in this sheet. If transform is True, the placement of
        the sheet will be added to each wire"""

        outp = []
        for p in obj.Group:
            ispanel = False
            if hasattr(p, "Proxy"):
                if hasattr(p.Proxy, "getWires"):
                    ispanel = True
                    w = p.Proxy.getWires(p)
                    if w[0]:
                        w = w[0]
                        w.scale(obj.Scale, FreeCAD.Vector())
                        if transform:
                            w.Placement = obj.Placement.multiply(w.Placement)
                        outp.append(w)
            if not ispanel:
                if hasattr(p, "Shape"):
                    for w in p.Shape.Wires:
                        w.scale(obj.Scale, FreeCAD.Vector())
                        if transform:
                            w.Placement = obj.Placement.multiply(w.Placement)
                        outp.append(w)
        return outp

    def getHoles(self, obj, transform=False):
        """getHoles(obj,transform=False): returns a list of compound whose wires define the
        holes contained in the panels in this sheet. If transform is True, the placement of
        the sheet will be added to each wire"""

        outp = []
        for p in obj.Group:
            if hasattr(p, "Proxy"):
                if hasattr(p.Proxy, "getWires"):
                    w = p.Proxy.getWires(p)
                    if w[1]:
                        w = w[1]
                        w.scale(obj.Scale, FreeCAD.Vector())
                        if transform:
                            w.Placement = obj.Placement.multiply(w.Placement)
                        outp.append(w)
        return outp

    def getTags(self, obj, transform=False):
        """getTags(obj,transform=False): returns a list of compounds whose wires define the
        tags (engravings) contained in the panels in this sheet and the sheet intself.
        If transform is True, the placement of the sheet will be added to each wire.
        Warning, the wires returned by this function may not be closed,
        depending on the font"""

        outp = []
        for p in obj.Group:
            if hasattr(p, "Proxy"):
                if hasattr(p.Proxy, "getWires"):
                    w = p.Proxy.getWires(p)
                    if w[2]:
                        w = w[2]
                        w.scale(obj.Scale, FreeCAD.Vector())
                        if transform:
                            w.Placement = obj.Placement.multiply(w.Placement)
                        outp.append(w)
        if self.sheettag is not None:
            w = self.sheettag.copy()
            w.scale(obj.Scale, FreeCAD.Vector())
            if transform:
                w.Placement = obj.Placement.multiply(w.Placement)
            outp.append(w)

        return outp


class ViewProviderPanelSheet(Draft.ViewProviderDraft):
    "a view provider for the panel sheet object"

    def __init__(self, vobj):

        Draft.ViewProviderDraft.__init__(self, vobj)
        self.setProperties(vobj)
        vobj.PatternSize = 0.0035

    def setProperties(self, vobj):

        pl = vobj.PropertiesList
        if not "Margin" in pl:
            vobj.addProperty(
                "App::PropertyLength",
                "Margin",
                "PanelSheet",
                QT_TRANSLATE_NOOP("App::Property", "A margin inside the boundary"),
                locked=True,
            )
        if not "ShowMargin" in pl:
            vobj.addProperty(
                "App::PropertyBool",
                "ShowMargin",
                "PanelSheet",
                QT_TRANSLATE_NOOP("App::Property", "Turns the display of the margin on/off"),
                locked=True,
            )
        if not "ShowGrain" in pl:
            vobj.addProperty(
                "App::PropertyBool",
                "ShowGrain",
                "PanelSheet",
                QT_TRANSLATE_NOOP(
                    "App::Property", "Turns the display of the wood grain texture on/off"
                ),
                locked=True,
            )

    def onDocumentRestored(self, vobj):

        self.setProperties(vobj)

    def getIcon(self):

        return ":/icons/Arch_Panel_Sheet_Tree.svg"

    def setEdit(self, vobj, mode):
        if mode == 1 or mode == 2:
            return None

        taskd = SheetTaskPanel(vobj.Object)
        taskd.update()
        FreeCADGui.Control.showDialog(taskd)
        return True

    def unsetEdit(self, vobj, mode):
        if mode == 1 or mode == 2:
            return None

        FreeCADGui.Control.closeDialog()
        return True

    def attach(self, vobj):

        Draft.ViewProviderDraft.attach(self, vobj)
        from pivy import coin

        self.coords = coin.SoCoordinate3()
        self.lineset = coin.SoLineSet()
        self.lineset.numVertices.setValue(-1)
        lineStyle = coin.SoDrawStyle()
        lineStyle.linePattern = 0x0F0F
        self.color = coin.SoBaseColor()
        self.switch = coin.SoSwitch()
        sep = coin.SoSeparator()
        self.switch.whichChild = -1
        sep.addChild(self.color)
        sep.addChild(lineStyle)
        sep.addChild(self.coords)
        sep.addChild(self.lineset)
        self.switch.addChild(sep)
        vobj.Annotation.addChild(self.switch)
        self.onChanged(vobj, "ShowMargin")
        self.onChanged(vobj, "LineColor")

    def onChanged(self, vobj, prop):

        if prop in ["Margin", "ShowMargin"]:
            if hasattr(vobj, "Margin") and hasattr(vobj, "ShowMargin"):
                if (
                    (vobj.Margin.Value > 0)
                    and (vobj.Margin.Value < vobj.Object.Width.Value / 2)
                    and (vobj.Margin.Value < vobj.Object.Height.Value / 2)
                ):
                    l2 = vobj.Object.Width.Value / 2
                    w2 = vobj.Object.Height.Value / 2
                    v = vobj.Margin.Value
                    v1 = (-l2 + v, -w2 + v, 0)
                    v2 = (l2 - v, -w2 + v, 0)
                    v3 = (l2 - v, w2 - v, 0)
                    v4 = (-l2 + v, w2 - v, 0)
                    self.coords.point.setValues([v1, v2, v3, v4, v1])
                    self.lineset.numVertices.setValue(5)
                if vobj.ShowMargin:
                    self.switch.whichChild = 0
                else:
                    self.switch.whichChild = -1
        elif prop == "LineColor":
            if hasattr(vobj, "LineColor"):
                c = vobj.LineColor
                self.color.rgb.setValue(c[0], c[1], c[2])
        elif prop == "ShowGrain":
            if hasattr(vobj, "ShowGrain"):
                if vobj.ShowGrain:
                    vobj.Pattern = "woodgrain"
                else:
                    vobj.Pattern = "None"
        Draft.ViewProviderDraft.onChanged(self, vobj, prop)

    def updateData(self, obj, prop):

        if prop in ["Width", "Height"]:
            self.onChanged(obj.ViewObject, "Margin")
        elif prop == "GrainDirection":
            if hasattr(self, "texcoords"):
                if self.texcoords:
                    s = FreeCAD.Vector(self.texcoords.directionS.getValue().getValue()).Length
                    vS = DraftVecUtils.rotate(
                        FreeCAD.Vector(s, 0, 0), -math.radians(obj.GrainDirection.Value)
                    )
                    vT = DraftVecUtils.rotate(
                        FreeCAD.Vector(0, s, 0), -math.radians(obj.GrainDirection.Value)
                    )
                    self.texcoords.directionS.setValue(vS.x, vS.y, vS.z)
                    self.texcoords.directionT.setValue(vT.x, vT.y, vT.z)
        Draft.ViewProviderDraft.updateData(self, obj, prop)


class SheetTaskPanel(ArchComponent.ComponentTaskPanel):

    def __init__(self, obj):

        ArchComponent.ComponentTaskPanel.__init__(self)
        self.obj = obj
        self.optwid = QtGui.QWidget()
        self.optwid.setWindowTitle(QtGui.QApplication.translate("Arch", "Tools", None))
        lay = QtGui.QVBoxLayout(self.optwid)
        self.editButton = QtGui.QPushButton(self.optwid)
        self.editButton.setIcon(QtGui.QIcon(":/icons/Draft_Edit.svg"))
        self.editButton.setText(QtGui.QApplication.translate("Arch", "Edit views positions", None))
        lay.addWidget(self.editButton)
        QtCore.QObject.connect(self.editButton, QtCore.SIGNAL("clicked()"), self.editNodes)
        self.form = [self.form, self.optwid]

    def editNodes(self):

        FreeCADGui.Control.closeDialog()
        FreeCADGui.runCommand("Draft_Edit")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2024 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM base task panel object"
__author__ = "Mario Passaglia"
__url__ = "https://www.freecad.org"

## @package base_femtaskpanel
#  \ingroup FEM
#  \brief base object for FEM task panels


class _BaseTaskPanel:
    """
    Base task panel
    """

    def __init__(self, obj):
        self.obj = obj

    def accept(self):
        gui_doc = self.obj.ViewObject.Document
        gui_doc.Document.recompute()
        gui_doc.resetEdit()
        gui_doc.Document.commitTransaction()

        return True

    def reject(self):
        gui_doc = self.obj.ViewObject.Document
        gui_doc.Document.abortTransaction()
        gui_doc.resetEdit()
        gui_doc.Document.recompute()

        return True



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
import Path
import Path.Op.Gui.Base as PathOpGui
import Path.Op.Vcarve as PathVcarve
import Path.Base.Gui.Util as PathGuiUtil

import PathGui
import PathScripts.PathUtils as PathUtils
from PySide import QtCore, QtGui


__title__ = "CAM Vcarve Operation UI"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Vcarve operation page controller and command implementation."

# There is a bug in logging library. To enable debugging - set True also in Op/Vcarve.py

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

translate = FreeCAD.Qt.translate


class TaskPanelBaseGeometryPage(PathOpGui.TaskPanelBaseGeometryPage):
    """Enhanced base geometry page to also allow special base objects."""

    def super(self):
        return super(TaskPanelBaseGeometryPage, self)

    def addBaseGeometry(self, selection):
        Path.Log.track(selection)
        added = False
        shapes = self.obj.BaseShapes
        for sel in selection:
            job = PathUtils.findParentJob(self.obj)
            base = job.Proxy.resourceClone(job, sel.Object)
            if not base:
                Path.Log.notice(
                    (translate("CAM", "%s is not a Base Model object of the job %s") + "\n")
                    % (sel.Object.Label, job.Label)
                )
                continue
            if base in shapes:
                Path.Log.notice("Base shape %s already in the list".format(sel.Object.Label))
                continue
            if base.isDerivedFrom("Part::Part2DObject"):
                if sel.HasSubObjects:
                    # selectively add some elements of the drawing to the Base
                    for sub in sel.SubElementNames:
                        if "Vertex" in sub:
                            Path.Log.info("Ignoring vertex")
                        else:
                            self.obj.Proxy.addBase(self.obj, base, sub)
                else:
                    # when adding an entire shape to BaseShapes we can take its sub shapes out of Base
                    self.obj.Base = [(p, el) for p, el in self.obj.Base if p != base]
                    shapes.append(base)
                    self.obj.BaseShapes = shapes
                added = True

        if not added:
            # user wants us to engrave an edge of face of a base model
            Path.Log.info("  call default")
            base = self.super().addBaseGeometry(selection)
            added = added or base

        return added

    def setFields(self, obj):
        self.super().setFields(obj)
        self.form.baseList.blockSignals(True)
        for shape in self.obj.BaseShapes:
            item = QtGui.QListWidgetItem(shape.Label)
            item.setData(self.super().DataObject, shape)
            item.setData(self.super().DataObjectSub, None)
            self.form.baseList.addItem(item)
        self.form.baseList.blockSignals(False)

    def updateBase(self):
        Path.Log.track()
        shapes = []
        for i in range(self.form.baseList.count()):
            item = self.form.baseList.item(i)
            obj = item.data(self.super().DataObject)
            sub = item.data(self.super().DataObjectSub)
            if not sub:
                shapes.append(obj)
        Path.Log.debug("Setting new base shapes: %s -> %s" % (self.obj.BaseShapes, shapes))
        self.obj.BaseShapes = shapes
        return self.super().updateBase()


class TaskPanelOpPage(PathOpGui.TaskPanelPage):
    """Page controller class for the Vcarve operation."""

    def initPage(self, obj):
        self.finishingPassZOffsetSpinBox = PathGuiUtil.QuantitySpinBox(
            self.form.finishingPassZOffset, obj, "FinishingPassZOffset"
        )

    def getForm(self):
        """getForm() ... returns UI"""
        form = FreeCADGui.PySideUic.loadUi(":/panels/PageOpVcarveEdit.ui")
        self.updateFormConditionalState(form)
        return form

    def updateFormConditionalState(self, form):
        """
        Update conditional form controls - i.e settings that should be
        visible only under certain conditions (other settings enabled, etc).
        """

        if form.finishingPassEnabled.isChecked():
            form.finishingPassZOffset.setVisible(True)
            form.finishingPassZOffsetLabel.setVisible(True)
        else:
            form.finishingPassZOffset.setVisible(False)
            form.finishingPassZOffsetLabel.setVisible(False)

    def updateFormCallback(self):
        return self.updateFormConditionalState(self.form)

    def registerSignalHandlers(self, obj):
        """Register signal handlers to update conditiona UI states"""

        self.form.finishingPassEnabled.stateChanged.connect(self.updateFormCallback)

    def getFields(self, obj):
        """getFields(obj) ... transfers values from UI to obj's properties"""
        if obj.Discretize != self.form.discretize.value():
            obj.Discretize = self.form.discretize.value()
        if obj.Colinear != self.form.colinearFilter.value():
            obj.Colinear = self.form.colinearFilter.value()

        if obj.FinishingPass != self.form.finishingPassEnabled.isChecked():
            obj.FinishingPass = self.form.finishingPassEnabled.isChecked()

        if obj.OptimizeMovements != self.form.optimizeMovementsEnabled.isChecked():
            obj.OptimizeMovements = self.form.optimizeMovementsEnabled.isChecked()

        self.finishingPassZOffsetSpinBox.updateProperty()

        self.updateCoolant(obj, self.form.coolantController)

        try:
            self.updateToolController(obj, self.form.toolController)
        except PathUtils.PathNoTCExistsException:
            title = translate("CAM", "No valid toolcontroller")
            message = translate(
                "CAM",
                "This operation requires a tool controller with a v-bit tool",
            )

            self.show_error_message(title, message)

    def setFields(self, obj):
        """setFields(obj) ... transfers obj's property values to UI"""
        self.form.discretize.setValue(obj.Discretize)
        self.form.colinearFilter.setValue(obj.Colinear)
        self.form.finishingPassEnabled.setChecked(obj.FinishingPass)
        self.form.optimizeMovementsEnabled.setChecked(obj.OptimizeMovements)

        self.finishingPassZOffsetSpinBox.updateWidget()

        self.setupToolController(obj, self.form.toolController)
        self.setupCoolant(obj, self.form.coolantController)

        self.updateFormConditionalState(self.form)

    def getSignalsForUpdate(self, obj):
        """getSignalsForUpdate(obj) ... return list of signals for updating obj"""
        signals = []
        signals.append(self.form.discretize.editingFinished)
        signals.append(self.form.colinearFilter.editingFinished)
        signals.append(self.form.finishingPassEnabled.stateChanged)
        signals.append(self.form.finishingPassZOffset.editingFinished)

        signals.append(self.form.optimizeMovementsEnabled.stateChanged)

        signals.append(self.form.toolController.currentIndexChanged)
        signals.append(self.form.coolantController.currentIndexChanged)
        return signals

    def taskPanelBaseGeometryPage(self, obj, features):
        """taskPanelBaseGeometryPage(obj, features) ... return page for adding base geometries."""
        return TaskPanelBaseGeometryPage(obj, features)


Command = PathOpGui.SetupOperation(
    "Vcarve",
    PathVcarve.Create,
    TaskPanelOpPage,
    "CAM_Vcarve",
    QtCore.QT_TRANSLATE_NOOP("CAM_Vcarve", "Vcarve"),
    QtCore.QT_TRANSLATE_NOOP("CAM_Vcarve", "Creates a medial line engraving toolpath"),
    PathVcarve.SetupProperties,
)

FreeCAD.Console.PrintLog("Loading PathVcarveGui... done\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>                  *
# *   (c) 2009, 2010 Ken Cline <cline@frii.com>                             *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to repair objects created with older versions."""
## @package gui_heal
# \ingroup draftguitools
# \brief Provides GUI tools to repair objects created with older versions.

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCADGui as Gui
import Draft
import draftguitools.gui_base as gui_base

from draftutils.translate import translate


class Heal(gui_base.GuiCommandSimplest):
    """The Draft Heal command definition.

    Heal faulty Draft objects saved with an earlier version of the program.

    It inherits `GuiCommandSimplest` to set up the document
    and other behavior. See this class for more information.
    """

    def __init__(self):
        super(Heal, self).__init__(name=translate("draft", "Heal"))

    def GetResources(self):
        """Set icon, menu and tooltip."""
        _tip = ()

        return {
            "Pixmap": "Draft_Heal",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Heal", "Heal"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Heal",
                "Heals faulty Draft objects saved with an earlier version of FreeCAD.\nIf an object is selected it tries to heal only that object,\notherwise it tries to heal all objects in the active document.",
            ),
        }

    def Activated(self):
        """Execute when the command is called."""
        super(Heal, self).Activated()

        s = Gui.Selection.getSelection()
        self.doc.openTransaction("Heal")
        if s:
            Draft.heal(s)
        else:
            Draft.heal()
        self.doc.commitTransaction()


Gui.addCommand("Draft_Heal", Heal())

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2024 Ondsel <development@ondsel.com>                    *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import os
from Path.Post.Processor import PostProcessor
import Path
import FreeCAD

Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

translate = FreeCAD.Qt.translate

debug = True
if debug:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class Generic(PostProcessor):
    def __init__(self, job):
        super().__init__(
            job,
            tooltip=translate("CAM", "Generic post processor"),
            tooltipargs=["arg1", "arg2"],
            units="kg",
        )
        Path.Log.debug("Generic post processor initialized")

    def export(self):
        Path.Log.debug("Exporting the job")

        postables = self._buildPostList()
        Path.Log.debug(f"postables count: {len(postables)}")

        g_code_sections = []
        for idx, section in enumerate(postables):
            partname, sublist = section

            # here is where the sections are converted to gcode.
            g_code_sections.append((idx, partname))

        return g_code_sections

    @property
    def tooltip(self):

        tooltip = """
        This is a generic post processor.
        It doesn't do anything yet because we haven't immplemented it.

        Implementing it would be a good idea
        """
        return tooltip

    @property
    def tooltipArgs(self):
        argtooltip = """
        --arg1: This is the first argument
        --arg2: This is the second argument

        """
        return argtooltip

    @property
    def units(self):
        return self._units



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Utility modules that are used throughout the workbench.

These modules include functions intended to be quite general,
so they should be useful in all other modules in the workbench
and even other workbenches.
Since they are not meant to be very complex, they shouldn't require
a lot of prerequisites, and shouldn't cause problems of circular imports.

They include modules that don't require the graphical interface (GUI),
as well as functions that do require it because they interact
with the view providers or with the 3D view.

Non GUI modules
---------------
- `utils`, generic functions
- `messages`, shorthands to print to the console
- `translate`, translate text using QtCore

GUI modules
-----------
- `gui_utils`, generic functions that deal with the graphical interface
- `todo`, delay execution of Python code through Qt

Initialization modules for the GUI
----------------------------------
- `init_tools`, initialize toolbars and menus of the workbench
- `init_draft_statusbar`, initialize the status bar of the workbench
"""
## \defgroup draftutils draftutils
# \ingroup DRAFT
# \brief Utility modules that are used throughout the workbench.



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2020 Carlo Pavan <carlopav@gmail.com>                             *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Modules that contain classes that define viewproviders for scripted objects.

These classes define viewproviders for the custom objects
defined in the `draftobjects` package.
They extend the basic viewprovider that is installed by default
with the creation of a base object defined in C++.

These classes are not normally used by themselves, but are used at creation
time by the make functions defined in the `draftmake` package.
These viewproviders are installed only when the graphical interface
is available; in console-only mode the viewproviders are not available
so the make functions ignore them.

Similar to the object classes, once a viewprovider class is assigned
to an object, and it is saved in a document, the object's viewprovider
will be rebuilt with the same class every time the document
is opened. In order to do this, the viewprovider module must exist
with the same name as when the object was saved.

For example, when creating a `Rectangle`, the object uses
the `draftviewproviders.view_rectangle.ViewProviderRectangle` class.
This class must exist when the document is opened again.

This means that, in general, these modules cannot be renamed or moved
without risking breaking previously saved files. They can be renamed
only if the old class can be migrated to point to a new class,
for example, by creating a reference to the new class named the same
as the older class.

::

    old_module.ViewProviderRectangle = new_module.ViewProviderRectangle
"""
## \defgroup draftviewproviders draftviewproviders
# \ingroup DRAFT
# \brief Classes that define viewproviders for the scripted objects.



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
import Path
import Path.Base.Gui.Util as PathGuiUtil
import Path.Op.Drilling as PathDrilling
import Path.Op.Gui.Base as PathOpGui
import Path.Op.Gui.CircularHoleBase as PathCircularHoleBaseGui
import PathGui

from PySide import QtCore

__title__ = "CAM Drilling Operation UI."
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "UI and Command for Drilling Operation."
__contributors__ = "IMBack!"

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class TaskPanelOpPage(PathCircularHoleBaseGui.TaskPanelOpPage):
    """Controller for the drilling operation's page"""

    def initPage(self, obj):
        self.peckDepthSpinBox = PathGuiUtil.QuantitySpinBox(self.form.peckDepth, obj, "PeckDepth")
        self.peckRetractSpinBox = PathGuiUtil.QuantitySpinBox(
            self.form.peckRetractHeight, obj, "RetractHeight"
        )
        self.dwellTimeSpinBox = PathGuiUtil.QuantitySpinBox(self.form.dwellTime, obj, "DwellTime")
        self.form.chipBreakEnabled.setEnabled(False)

    def registerSignalHandlers(self, obj):
        self.form.peckEnabled.toggled.connect(self.form.peckDepth.setEnabled)
        self.form.peckEnabled.toggled.connect(self.form.dwellEnabled.setDisabled)
        self.form.peckEnabled.toggled.connect(self.form.feedRetractEnabled.setDisabled)
        self.form.peckEnabled.toggled.connect(self.setChipBreakControl)

        self.form.feedRetractEnabled.toggled.connect(self.form.peckDepth.setDisabled)
        self.form.feedRetractEnabled.toggled.connect(self.form.peckEnabled.setDisabled)
        self.form.feedRetractEnabled.toggled.connect(self.form.dwellEnabled.setDisabled)
        self.form.feedRetractEnabled.toggled.connect(self.form.chipBreakEnabled.setDisabled)
        self.form.feedRetractEnabled.toggled.connect(self.form.peckEnabled.setDisabled)
        self.form.feedRetractEnabled.toggled.connect(self.setChipBreakControl)

        self.form.dwellEnabled.toggled.connect(self.form.dwellTime.setEnabled)
        self.form.dwellEnabled.toggled.connect(self.form.dwellTimelabel.setEnabled)
        self.form.dwellEnabled.toggled.connect(self.form.peckEnabled.setDisabled)
        self.form.dwellEnabled.toggled.connect(self.form.feedRetractEnabled.setDisabled)
        self.form.dwellEnabled.toggled.connect(self.setChipBreakControl)

        self.form.peckRetractHeight.setEnabled(True)
        self.form.retractLabel.setEnabled(True)

        if self.form.peckEnabled.isChecked():
            self.form.dwellEnabled.setEnabled(False)
            self.form.feedRetractEnabled.setEnabled(False)
            self.form.peckDepth.setEnabled(True)
            self.form.peckDepthLabel.setEnabled(True)
            self.form.chipBreakEnabled.setEnabled(True)
        elif self.form.dwellEnabled.isChecked():
            self.form.feedRetractEnabled.setEnabled(False)
            self.form.peckEnabled.setEnabled(False)
            self.form.dwellTime.setEnabled(True)
            self.form.dwellTimelabel.setEnabled(True)
            self.form.chipBreakEnabled.setEnabled(False)
        else:
            self.form.feedRetractEnabled.setEnabled(True)
            self.form.chipBreakEnabled.setEnabled(False)

    def setChipBreakControl(self):
        self.form.chipBreakEnabled.setEnabled(self.form.peckEnabled.isChecked())

    def getForm(self):
        """getForm() ... return UI"""
        form = FreeCADGui.PySideUic.loadUi(":/panels/PageOpDrillingEdit.ui")

        comboToPropertyMap = [("ExtraOffset", "ExtraOffset")]
        enumTups = PathDrilling.ObjectDrilling.propertyEnumerations(dataType="raw")
        self.populateCombobox(form, enumTups, comboToPropertyMap)

        return form

    def updateQuantitySpinBoxes(self, index=None):
        self.peckDepthSpinBox.updateWidget()
        self.peckRetractSpinBox.updateWidget()
        self.dwellTimeSpinBox.updateWidget()

    def getFields(self, obj):
        """setFields(obj) ... update obj's properties with values from the UI"""
        Path.Log.track()
        self.peckDepthSpinBox.updateProperty()
        self.peckRetractSpinBox.updateProperty()
        self.dwellTimeSpinBox.updateProperty()

        if obj.KeepToolDown != self.form.KeepToolDownEnabled.isChecked():
            obj.KeepToolDown = self.form.KeepToolDownEnabled.isChecked()
        if obj.DwellEnabled != self.form.dwellEnabled.isChecked():
            obj.DwellEnabled = self.form.dwellEnabled.isChecked()
        if obj.PeckEnabled != self.form.peckEnabled.isChecked():
            obj.PeckEnabled = self.form.peckEnabled.isChecked()
        if obj.feedRetractEnabled != self.form.feedRetractEnabled.isChecked():
            obj.feedRetractEnabled = self.form.feedRetractEnabled.isChecked()
        if obj.chipBreakEnabled != self.form.chipBreakEnabled.isChecked():
            obj.chipBreakEnabled = self.form.chipBreakEnabled.isChecked()
        if obj.ExtraOffset != str(self.form.ExtraOffset.currentData()):
            obj.ExtraOffset = str(self.form.ExtraOffset.currentData())

        self.updateToolController(obj, self.form.toolController)
        self.updateCoolant(obj, self.form.coolantController)

    def setFields(self, obj):
        """setFields(obj) ... update UI with obj properties' values"""
        Path.Log.track()
        self.updateQuantitySpinBoxes()

        if not hasattr(obj, "KeepToolDown"):
            obj.addProperty(
                "App::PropertyBool",
                "KeepToolDown",
                "Drill",
                QtCore.QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Apply G99 retraction: only retract to RetractHeight between holes in this operation",
                ),
            )

        if obj.KeepToolDown:
            self.form.KeepToolDownEnabled.setCheckState(QtCore.Qt.Checked)
        else:
            self.form.KeepToolDownEnabled.setCheckState(QtCore.Qt.Unchecked)

        if obj.DwellEnabled:
            self.form.dwellEnabled.setCheckState(QtCore.Qt.Checked)
        else:
            self.form.dwellEnabled.setCheckState(QtCore.Qt.Unchecked)

        if obj.PeckEnabled:
            self.form.peckEnabled.setCheckState(QtCore.Qt.Checked)
        else:
            self.form.peckEnabled.setCheckState(QtCore.Qt.Unchecked)
            self.form.chipBreakEnabled.setEnabled(False)

        if obj.chipBreakEnabled:
            self.form.chipBreakEnabled.setCheckState(QtCore.Qt.Checked)
        else:
            self.form.chipBreakEnabled.setCheckState(QtCore.Qt.Unchecked)

        if obj.feedRetractEnabled:
            self.form.feedRetractEnabled.setCheckState(QtCore.Qt.Checked)
        else:
            self.form.feedRetractEnabled.setCheckState(QtCore.Qt.Unchecked)

        self.selectInComboBox(obj.ExtraOffset, self.form.ExtraOffset)

        self.setupToolController(obj, self.form.toolController)
        self.setupCoolant(obj, self.form.coolantController)

    def getSignalsForUpdate(self, obj):
        """getSignalsForUpdate(obj) ... return list of signals which cause the receiver to update the model"""
        signals = []

        signals.append(self.form.peckRetractHeight.editingFinished)
        signals.append(self.form.peckDepth.editingFinished)
        signals.append(self.form.dwellTime.editingFinished)
        if hasattr(self.form.dwellEnabled, "checkStateChanged"):  # Qt version >= 6.7.0
            signals.append(self.form.dwellEnabled.checkStateChanged)
            signals.append(self.form.peckEnabled.checkStateChanged)
            signals.append(self.form.chipBreakEnabled.checkStateChanged)
        else:  # Qt version < 6.7.0
            signals.append(self.form.dwellEnabled.stateChanged)
            signals.append(self.form.peckEnabled.stateChanged)
            signals.append(self.form.chipBreakEnabled.stateChanged)
        signals.append(self.form.toolController.currentIndexChanged)
        signals.append(self.form.coolantController.currentIndexChanged)
        signals.append(self.form.ExtraOffset.currentIndexChanged)
        if hasattr(self.form.KeepToolDownEnabled, "checkStateChanged"):  # Qt version >= 6.7.0
            signals.append(self.form.KeepToolDownEnabled.checkStateChanged)
            signals.append(self.form.feedRetractEnabled.checkStateChanged)
        else:  # Qt version < 6.7.0
            signals.append(self.form.KeepToolDownEnabled.stateChanged)
            signals.append(self.form.feedRetractEnabled.stateChanged)
        return signals

    def updateData(self, obj, prop):
        if prop in ["PeckDepth", "RetractHeight"] and not prop in ["Base", "Disabled"]:
            self.updateQuantitySpinBoxes()


Command = PathOpGui.SetupOperation(
    "Drilling",
    PathDrilling.Create,
    TaskPanelOpPage,
    "CAM_Drilling",
    QtCore.QT_TRANSLATE_NOOP("CAM_Drilling", "Drilling"),
    QtCore.QT_TRANSLATE_NOOP(
        "CAM_Drilling",
        "Creates a Drilling toolpath from the features of a base object",
    ),
    PathDrilling.SetupProperties,
)

FreeCAD.Console.PrintLog("Loading PathDrillingGui... done\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
import Path
import Path.Base.Gui.GetPoint as PathGetPoint
import Path.Base.Gui.Util as PathGuiUtil
import Path.Base.SetupSheet as PathSetupSheet
import Path.Base.Util as PathUtil
import Path.Main.Job as PathJob
import Path.Op.Base as PathOp
import Path.Op.Gui.Selection as PathSelection
import Path.Tool.Controller as PathToolController
from Path.Tool.library.ui.dock import ToolBitLibraryDock
import PathScripts.PathUtils as PathUtils
import importlib
from PySide.QtCore import QT_TRANSLATE_NOOP

from PySide import QtCore, QtGui, QtWidgets

__title__ = "CAM Operation UI base classes"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Base classes and framework for CAM operation's UI"

translate = FreeCAD.Qt.translate

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class ViewProvider(object):
    """
    Generic view provider for path objects.
    Deducts the icon name from operation name, brings up the TaskPanel
    with pages corresponding to the operation's opFeatures() and forwards
    property change notifications to the page controllers.
    """

    def __init__(self, vobj, resources):
        Path.Log.track()
        self.deleteOnReject = True
        self.OpIcon = ":/icons/%s.svg" % resources.pixmap
        self.OpName = resources.name
        self.OpPageModule = resources.opPageClass.__module__
        self.OpPageClass = resources.opPageClass.__name__

        # initialized later
        self.vobj = vobj
        self.Object = None
        self.panel = None

    def attach(self, vobj):
        Path.Log.track()
        self.vobj = vobj
        self.Object = vobj.Object
        self.panel = None
        return

    def deleteObjectsOnReject(self):
        """
        deleteObjectsOnReject() ... return true if all objects should
        be created if the user hits cancel. This is used during the initial
        edit session, if the user does not press OK, it is assumed they've
        changed their mind about creating the operation.
        """
        Path.Log.track()
        return hasattr(self, "deleteOnReject") and self.deleteOnReject

    def setDeleteObjectsOnReject(self, state=False):
        Path.Log.track()
        self.deleteOnReject = state
        return self.deleteOnReject

    def setEdit(self, vobj=None, mode=0):
        """setEdit(vobj, mode=0) ... initiate editing of receivers model."""
        Path.Log.track()
        if 0 == mode:
            if vobj is None:
                vobj = self.vobj
            page = self.getTaskPanelOpPage(vobj.Object)
            page.setTitle(self.OpName)
            page.setIcon(self.OpIcon)
            selection = self.getSelectionFactory()
            self.setupTaskPanel(
                TaskPanel(vobj.Object, self.deleteObjectsOnReject(), page, selection)
            )
            self.deleteOnReject = False
            return True
        # no other editing possible
        return False

    def setupTaskPanel(self, panel):
        """setupTaskPanel(panel) ... internal function to start the editor."""
        self.panel = panel
        FreeCADGui.Control.closeDialog()
        FreeCADGui.Control.showDialog(panel)
        panel.setupUi()
        job = self.Object.Proxy.getJob(self.Object)
        if job:
            job.ViewObject.Proxy.setupEditVisibility(job)
        else:
            Path.Log.info("did not find no job")

    def clearTaskPanel(self):
        """clearTaskPanel() ... internal callback function when editing has finished."""
        self.panel = None
        job = self.Object.Proxy.getJob(self.Object)
        if job:
            job.ViewObject.Proxy.resetEditVisibility(job)

    def unsetEdit(self, arg1, arg2):
        if self.panel:
            self.panel.reject(False)

    def dumps(self):
        """dumps() ... callback before receiver is saved to a file.
        Returns a dictionary with the receiver's resources as strings."""
        Path.Log.track()
        state = {}
        state["OpName"] = self.OpName
        state["OpIcon"] = self.OpIcon
        state["OpPageModule"] = self.OpPageModule
        state["OpPageClass"] = self.OpPageClass
        return state

    def loads(self, state):
        """loads(state) ... callback on restoring a saved instance, pendant to dumps()
        state is the dictionary returned by dumps()."""
        self.OpName = state["OpName"]
        self.OpIcon = state["OpIcon"]
        self.OpPageModule = state["OpPageModule"]
        self.OpPageClass = state["OpPageClass"]

    def getIcon(self):
        """getIcon() ... the icon used in the object tree"""
        if self.Object.Active:
            return self.OpIcon
        else:
            return ":/icons/CAM_OpActive.svg"

    def getTaskPanelOpPage(self, obj):
        """getTaskPanelOpPage(obj) ... use the stored information to instantiate the receiver op's page controller."""
        mod = importlib.import_module(self.OpPageModule)
        cls = getattr(mod, self.OpPageClass)
        return cls(obj, 0)

    def getSelectionFactory(self):
        """getSelectionFactory() ... return a factory function that can be used to create the selection observer."""
        return PathSelection.select(self.OpName)

    def updateData(self, obj, prop):
        """updateData(obj, prop) ... callback whenever a property of the receiver's model is assigned.
        The callback is forwarded to the task panel - in case an editing session is ongoing."""
        # Path.Log.track(obj.Label, prop) # Creates a lot of noise
        if self.panel:
            self.panel.updateData(obj, prop)

    def onDelete(self, vobj, arg2=None):
        PathUtil.clearExpressionEngine(vobj.Object)
        return True

    def setupContextMenu(self, vobj, menu):
        Path.Log.track()
        for action in menu.actions():
            menu.removeAction(action)
        action = QtGui.QAction(translate("PathOp", "Edit"), menu)
        action.triggered.connect(self._editInContextMenuTriggered)
        menu.addAction(action)

    def _editInContextMenuTriggered(self, checked):
        self.setEdit()


class TaskPanelPage(object):
    """Base class for all task panel pages."""

    # task panel interaction framework
    def __init__(self, obj, features):
        """__init__(obj, features) ... framework initialisation.
        Do not overwrite, implement initPage(obj) instead."""
        self.obj = obj
        self.job = PathUtils.findParentJob(obj)
        self.form = self.getForm()
        self.signalDirtyChanged = None
        self.setClean()
        self.setTitle("-")
        self.setIcon(None)
        self.features = features
        self.isdirty = False
        self.parent = None
        self.panelTitle = "Operation"
        self.tcEditor = None
        self.combo = None

        if self._installTCUpdate():
            PathJob.Notification.updateTC.connect(self.resetToolController)
            self.form.toolController.currentIndexChanged.connect(self.tcComboChanged)

    def show_error_message(self, title, message):
        msg_box = QtGui.QMessageBox()
        msg_box.setIcon(QtGui.QMessageBox.Critical)
        msg_box.setWindowTitle(title)
        msg_box.setText(message)
        msg_box.setStandardButtons(QtGui.QMessageBox.Ok)
        msg_box.exec_()

    def _installTCUpdate(self):
        return hasattr(self.form, "toolController")

    def setParent(self, parent):
        """setParent() ... used to transfer parent object link to child class.
        Do not overwrite."""
        self.parent = parent

    def onDirtyChanged(self, callback):
        """onDirtyChanged(callback) ... set callback when dirty state changes."""
        self.signalDirtyChanged = callback

    def setDirty(self):
        """setDirty() ... mark receiver as dirty, causing the model to be recalculated if OK or Apply is pressed."""
        self.isdirty = True
        if self.signalDirtyChanged:
            self.signalDirtyChanged(self)

    def setClean(self):
        """setClean() ... mark receiver as clean, indicating there is no need to recalculate the model even if the user presses OK or Apply."""
        self.isdirty = False
        if self.signalDirtyChanged:
            self.signalDirtyChanged(self)

    def pageGetFields(self):
        """pageGetFields() ... internal callback.
        Do not overwrite, implement getFields(obj) instead."""
        self.getFields(self.obj)
        self.setDirty()

    def pageSetFields(self):
        """pageSetFields() ... internal callback.
        Do not overwrite, implement setFields(obj) instead."""
        self.setFields(self.obj)

    def pageCleanup(self):
        """pageCleanup() ... internal callback.
        Do not overwrite, implement cleanupPage(obj) instead."""
        if self._installTCUpdate():
            PathJob.Notification.updateTC.disconnect(self.resetToolController)
        self.cleanupPage(self.obj)

    def pageRegisterSignalHandlers(self):
        """pageRegisterSignalHandlers() .. internal callback.
        Registers a callback for all signals returned by getSignalsForUpdate(obj).
        Do not overwrite, implement getSignalsForUpdate(obj) and/or registerSignalHandlers(obj) instead.
        """
        for signal in self.getSignalsForUpdate(self.obj):
            signal.connect(self.pageGetFields)
        self.registerSignalHandlers(self.obj)

    def pageUpdateData(self, obj, prop):
        """pageUpdateData(obj, prop) ... internal callback.
        Do not overwrite, implement updateData(obj) instead."""
        self.updateData(obj, prop)

    def setTitle(self, title):
        """setTitle(title) ... sets a title for the page."""
        self.title = title

    def getTitle(self, obj):
        """getTitle(obj) ... return title to be used for the receiver page.
        The default implementation returns what was previously set with setTitle(title).
        Can safely be overwritten by subclasses."""
        return self.title

    def setIcon(self, icon):
        """setIcon(icon) ... sets the icon for the page."""
        self.icon = icon

    def getIcon(self, obj):
        """getIcon(obj) ... return icon for page or None.
        Can safely be overwritten by subclasses."""
        return self.icon

    # subclass interface
    def initPage(self, obj):
        """initPage(obj) ... overwrite to customize UI for specific model.
        Note that this function is invoked after all page controllers have been created.
        Should be overwritten by subclasses."""
        pass

    def cleanupPage(self, obj):
        """cleanupPage(obj) ... overwrite to perform any cleanup tasks before page is destroyed.
        Can safely be overwritten by subclasses."""
        pass

    def modifyStandardButtons(self, buttonBox):
        """modifyStandardButtons(buttonBox) ... overwrite if the task panel standard buttons need to be modified.
        Can safely be overwritten by subclasses."""
        pass

    def getForm(self):
        """getForm() ... return UI form for this page.
        Must be overwritten by subclasses."""
        pass

    def getFields(self, obj):
        """getFields(obj) ... overwrite to transfer values from UI to obj's properties.
        Can safely be overwritten by subclasses."""
        pass

    def setFields(self, obj):
        """setFields(obj) ... overwrite to transfer obj's property values to UI.
        Can safely be overwritten by subclasses."""
        pass

    def getSignalsForUpdate(self, obj):
        """getSignalsForUpdate(obj) ... return signals which, when triggered, cause the receiver to update the model.
        See also registerSignalHandlers(obj)
        Can safely be overwritten by subclasses."""
        return []

    def registerSignalHandlers(self, obj):
        """registerSignalHandlers(obj) ... overwrite to register custom signal handlers.
        In case an update of a model is not the desired operation of a signal invocation
        (see getSignalsForUpdate(obj)) this function can be used to register signal handlers
        manually.
        Can safely be overwritten by subclasses."""
        pass

    def updateData(self, obj, prop):
        """updateData(obj, prop) ... overwrite if the receiver needs to react to property changes that might not have been caused by the receiver itself.
        Sometimes a model will recalculate properties based on a change of another property. In order to keep the UI up to date with such changes this
        function can be used.
        Please note that the callback is synchronous with the property assignment operation. Also note that the notification is invoked regardless of the
        actual value of the property assignment. In other words it also fires if a property gets assigned the same value it already has.
        Taking above observations into account the implementation has to take care that it doesn't overwrite modified UI values by invoking setFields(obj).
        This can happen if a subclass unconditionally transfers all values in getFields(obj) to the model and just calls setFields(obj) in this callback.
        In such a scenario the first property assignment will cause all changes in the UI of the other fields to be overwritten by setFields(obj).
        You have been warned."""
        pass

    def updateSelection(self, obj, sel):
        """updateSelection(obj, sel) ...
        overwrite to customize UI depending on current selection.
        Can safely be overwritten by subclasses."""
        pass

    def selectInComboBox(self, name, combo):
        """selectInComboBox(name, combo) ...
        helper function to select a specific value in a combo box."""
        try:
            combo.blockSignals(True)
            index = combo.currentIndex()  # Save initial index

            # Search using currentData and return if found
            newindex = combo.findData(name)
            if newindex >= 0:
                combo.setCurrentIndex(newindex)
                return

            # if not found, search using current text
            newindex = combo.findText(name, QtCore.Qt.MatchFixedString)
            if newindex >= 0:
                combo.setCurrentIndex(newindex)
                return

            # not found, return unchanged
            combo.setCurrentIndex(index)
        finally:
            combo.blockSignals(False)

    def populateCombobox(self, form, enumTups, comboBoxesPropertyMap):
        """populateCombobox(form, enumTups, comboBoxesPropertyMap) ... proxy for PathGuiUtil.populateCombobox()"""
        PathGuiUtil.populateCombobox(form, enumTups, comboBoxesPropertyMap)

    def tcComboChanged(self, newIndex):
        if self.obj is not None and self.tcEditor:
            if newIndex == self.combo.count() - 1:
                # Special entry: new tool controller. Show the tool dock and reset combo
                dock = ToolBitLibraryDock(self.job, True)
                dock.open()
                self.resetTCCombo()
            elif newIndex == self.combo.count() - 2:
                # Special entry: copy tool controller
                self.copyToolController()  # this function also rebuilds the combo
                self.resetTCCombo()
            else:
                tc = PathUtils.findToolController(
                    self.obj, self.obj.Proxy, self.form.toolController.currentText()
                )
                self.obj.ToolController = tc
                self.setupToolController()

    def updateToolControllerEditorVisibility(self):
        if self.form.editToolController.isChecked():
            self.tcEditor.controller.show()
        else:
            self.tcEditor.controller.hide()

    def resetToolController(self, job, tc):
        if self.obj is None:
            return
        self.obj.ToolController = tc
        self.setupToolController()

    def copyToolController(self):
        oldTc = self.tcEditor.obj
        self.tcEditor.updateToolController()
        tc = PathToolController.Create(
            name=oldTc.Label, tool=oldTc.Tool, toolNumber=oldTc.ToolNumber
        )
        job = self.obj.Proxy.getJob(self.obj)
        job.Proxy.addToolController(tc)

        tc.HorizFeed = oldTc.HorizFeed
        tc.VertFeed = oldTc.VertFeed
        tc.HorizRapid = oldTc.HorizRapid
        tc.VertRapid = oldTc.VertRapid
        tc.SpindleSpeed = oldTc.SpindleSpeed
        tc.SpindleDir = oldTc.SpindleDir
        for attr, expr in oldTc.ExpressionEngine:
            tc.setExpression(attr, expr)

        self.obj.ToolController = tc
        self.setupToolController()

    def tcEditorChanged(self):
        self.setDirty()
        self.resetTCCombo()

    def resetTCCombo(self):
        controllers = PathUtils.getToolControllers(self.obj)

        if self.obj.ToolController is None:
            self.obj.ToolController = PathUtils.findToolController(self.obj, self.obj.Proxy)
        if len(controllers) > 0 and not self.obj.Proxy.isToolSupported(
            self.obj, self.obj.ToolController.Tool
        ):
            self.obj.ToolController = controllers[0]

        tcName = self.obj.ToolController.Label if self.obj.ToolController else ""
        labels = [c.Label for c in controllers]
        labels.append(FreeCAD.Qt.translate("CAM_Operation", "Copy {0}…").format(tcName))
        labels.append(FreeCAD.Qt.translate("CAM_Operation", "New tool controller…"))
        self.combo.blockSignals(True)
        self.combo.clear()
        self.combo.addItems(labels)
        self.combo.insertSeparator(len(controllers))
        self.combo.blockSignals(False)

        if self.obj.ToolController is not None:
            self.selectInComboBox(self.obj.ToolController.Label, self.combo)

    def setupToolController(self, obj=None, combo=None):
        """setupToolController(obj, combo) ...
        helper function to setup obj's ToolController
        in the given combo box."""
        obj = obj or self.obj
        combo = combo or self.combo
        self.obj, self.combo = obj, combo

        self.resetTCCombo()

        if hasattr(self.form, "editToolController"):
            layout = self.form.editToolController.parent().layout()
            oldEditor = self.tcEditor

            # Count the number of times the tool controller is used in other operations
            # If it is used in other operations, we will offer the "copy tool controller" button
            tcCount = 0
            for job in PathUtils.GetJobs():
                for op in job.Operations.Group:
                    if op == self.obj:
                        continue
                    elif hasattr(op, "ToolController") and op.ToolController == obj.ToolController:
                        tcCount += 1

            self.tcEditor = Path.Tool.Gui.Controller.ToolControllerEditor(
                obj.ToolController,
                False,
                self.tcEditorChanged,
                True,
                True,
            )
            self.tcEditor.setupUi()

            labelStr = FreeCAD.Qt.translate(
                "CAM_Operation", "This tool controller is used by {0} other operations."
            ).format(tcCount)
            self.tcEditor.controller.tcOperationCountLabel.setText(labelStr)

            # add to layout -- requires a grid layout
            if isinstance(layout, QtWidgets.QGridLayout):
                layout.addWidget(
                    self.tcEditor.controller, layout.rowCount(), 0, 1, layout.columnCount()
                )
            else:
                Path.Log.error(
                    "Panel uses a layout incompatible with editing tool controllers. Report a bug: it should be a QGridLayout"
                )

            self.updateToolControllerEditorVisibility()
            self.tcEditor.updateUi()
            checkbox = self.form.editToolController
            checkboxSignal = (
                checkbox.checkStateChanged
                if hasattr(checkbox, "checkStateChanged")
                else checkbox.stateChanged
            )
            checkboxSignal.connect(self.updateToolControllerEditorVisibility)

            if oldEditor:
                oldEditor.updateToolController()
                oldEditor.controller.hide()
                layout.removeWidget(oldEditor.controller)

    def updateToolController(self, obj, combo):
        """updateToolController(obj, combo) ...
        helper function to update obj's ToolController property if a different
        one has been selected in the combo box."""
        tc = PathUtils.findToolController(obj, obj.Proxy, combo.currentText())
        if obj.ToolController != tc:
            obj.ToolController = tc
        if self.tcEditor:
            self.tcEditor.updateToolController()

    def setupCoolant(self, obj, combo):
        """setupCoolant(obj, combo) ...
        helper function to setup obj's Coolant option."""
        job = PathUtils.findParentJob(obj)
        options = job.SetupSheet.CoolantModes
        combo.blockSignals(True)
        combo.clear()
        combo.addItems(options)
        combo.blockSignals(False)

        if hasattr(obj, "CoolantMode"):
            self.selectInComboBox(obj.CoolantMode, combo)

    def updateCoolant(self, obj, combo):
        """updateCoolant(obj, combo) ...
        helper function to update obj's Coolant property if a different
        one has been selected in the combo box."""
        option = combo.currentText()
        if hasattr(obj, "CoolantMode"):
            if obj.CoolantMode != option:
                obj.CoolantMode = option

    def updatePanelVisibility(self, panelTitle, obj):
        """updatePanelVisibility(panelTitle, obj) ...
        Function to call the `updateVisibility()` GUI method of the
        page whose panel title is as indicated."""
        if hasattr(self, "parent"):
            parent = getattr(self, "parent")
            if parent and hasattr(parent, "featurePages"):
                for page in parent.featurePages:
                    if hasattr(page, "panelTitle"):
                        if page.panelTitle == panelTitle and hasattr(page, "updateVisibility"):
                            page.updateVisibility()
                            break


class TaskPanelBaseGeometryPage(TaskPanelPage):
    """Page controller for the base geometry."""

    DataObject = QtCore.Qt.ItemDataRole.UserRole
    DataObjectSub = QtCore.Qt.ItemDataRole.UserRole + 1

    def __init__(self, obj, features):
        super(TaskPanelBaseGeometryPage, self).__init__(obj, features)

        self.panelTitle = "Base Geometry"
        self.OpIcon = ":/icons/CAM_BaseGeometry.svg"
        self.setIcon(self.OpIcon)

    def getForm(self):
        panel = FreeCADGui.PySideUic.loadUi(":/panels/PageBaseGeometryEdit.ui")
        self.modifyPanel(panel)
        return panel

    def modifyPanel(self, panel):
        """modifyPanel(self, panel) ...
        Helper method to modify the current form immediately after
        it is loaded."""
        # Determine if Job operations are available with Base Geometry
        availableOps = list()
        ops = self.job.Operations.Group
        for op in ops:
            if hasattr(op, "Base") and isinstance(op.Base, list):
                if len(op.Base) > 0:
                    availableOps.append(op.Label)

        # Load available operations into combobox
        if len(availableOps) > 0:
            # Populate the operations list
            try:
                panel.geometryImportList.blockSignals(True)
                panel.geometryImportList.clear()
                availableOps.sort()
                for opLbl in availableOps:
                    panel.geometryImportList.addItem(opLbl)
                panel.geometryImportList.blockSignals(False)
            except (AttributeError, RuntimeError):
                # Widget doesn't exist in UI or C++ object already deleted
                pass
        else:
            try:
                panel.geometryImportList.hide()
                panel.geometryImportButton.hide()
            except (AttributeError, RuntimeError):
                # Widget doesn't exist in UI or C++ object already deleted
                pass

    def getTitle(self, obj):
        return translate("PathOp", "Base Geometry")

    def getFields(self, obj):
        pass

    def setFields(self, obj):
        self.form.baseList.blockSignals(True)
        self.form.baseList.clear()
        for base in self.obj.Base:
            for sub in base[1]:
                item = QtGui.QListWidgetItem("%s.%s" % (base[0].Label, sub))
                item.setData(self.DataObject, base[0])
                item.setData(self.DataObjectSub, sub)
                self.form.baseList.addItem(item)
        self.form.baseList.blockSignals(False)
        self.resizeBaseList()

    def itemActivated(self):
        FreeCADGui.Selection.clearSelection()
        for item in self.form.baseList.selectedItems():
            obj = item.data(self.DataObject)
            sub = item.data(self.DataObjectSub)
            if sub:
                FreeCADGui.Selection.addSelection(obj, sub)
            else:
                FreeCADGui.Selection.addSelection(obj)
        # FreeCADGui.updateGui()

    def supportsVertexes(self):
        return self.features & PathOp.FeatureBaseVertexes

    def supportsEdges(self):
        return self.features & PathOp.FeatureBaseEdges

    def supportsFaces(self):
        return self.features & PathOp.FeatureBaseFaces

    def supportsPanels(self):
        return self.features & PathOp.FeatureBasePanels

    def featureName(self):
        if self.supportsEdges() and self.supportsFaces():
            return "features"
        if self.supportsFaces():
            return "faces"
        if self.supportsEdges():
            return "edges"
        return "nothing"

    def selectionSupportedAsBaseGeometry(self, selection, ignoreErrors):
        if len(selection) != 1:
            if not ignoreErrors:
                msg = translate(
                    "PathOp",
                    "Please select %s from a single solid" % self.featureName(),
                )
                Path.Log.debug(msg)
            return False
        sel = selection[0]
        if sel.HasSubObjects:
            if not self.supportsVertexes() and selection[0].SubObjects[0].ShapeType == "Vertex":
                return False
            if not self.supportsEdges() and selection[0].SubObjects[0].ShapeType == "Edge":
                return False
            if not self.supportsFaces() and selection[0].SubObjects[0].ShapeType == "Face":
                return False
        else:
            if not self.supportsPanels() or "Panel" not in sel.Object.Name:
                return False
        return True

    def addBaseGeometry(self, selection):
        Path.Log.track(selection)
        if self.selectionSupportedAsBaseGeometry(selection, False):
            sel = selection[0]
            for sub in sel.SubElementNames:
                self.obj.Proxy.addBase(self.obj, sel.Object, sub)
            return True
        return False

    def addBase(self):
        Path.Log.track()
        if self.addBaseGeometry(FreeCADGui.Selection.getSelectionEx()):
            # self.obj.Proxy.execute(self.obj)
            self.setFields(self.obj)
            self.setDirty()
            self.updatePanelVisibility("Operation", self.obj)

    def deleteBase(self):
        Path.Log.track()
        selected = self.form.baseList.selectedItems()
        for item in selected:
            self.form.baseList.takeItem(self.form.baseList.row(item))
            self.setDirty()
        self.updateBase()
        self.updatePanelVisibility("Operation", self.obj)
        self.resizeBaseList()

    def updateBase(self):
        newlist = []
        for i in range(self.form.baseList.count()):
            item = self.form.baseList.item(i)
            obj = item.data(self.DataObject)
            sub = item.data(self.DataObjectSub)
            if sub:
                base = (obj, str(sub))
                newlist.append(base)
        Path.Log.debug("Setting new base: %s -> %s" % (self.obj.Base, newlist))
        self.obj.Base = newlist

    def clearBase(self):
        self.obj.Base = []
        self.setDirty()
        self.updatePanelVisibility("Operation", self.obj)
        self.resizeBaseList()

    def importBaseGeometry(self):
        opLabel = str(self.form.geometryImportList.currentText())
        ops = FreeCAD.ActiveDocument.getObjectsByLabel(opLabel)
        if len(ops) > 1:
            msg = translate("PathOp", "Multiple operations are labeled as")
            msg += " {}\n".format(opLabel)
            FreeCAD.Console.PrintWarning(msg)
        for base, subList in ops[0].Base:
            FreeCADGui.Selection.clearSelection()
            FreeCADGui.Selection.addSelection(base, subList)
            self.addBase()

    def registerSignalHandlers(self, obj):
        self.form.baseList.itemSelectionChanged.connect(self.itemActivated)
        self.form.addBase.clicked.connect(self.addBase)
        self.form.deleteBase.clicked.connect(self.deleteBase)
        self.form.clearBase.clicked.connect(self.clearBase)
        self.form.geometryImportButton.clicked.connect(self.importBaseGeometry)

    def pageUpdateData(self, obj, prop):
        if prop in ["Base"]:
            self.setFields(obj)

    def updateSelection(self, obj, sel):
        if self.selectionSupportedAsBaseGeometry(sel, True):
            self.form.addBase.setEnabled(True)
        else:
            self.form.addBase.setEnabled(False)

    def resizeBaseList(self):
        # Set base geometry list window to resize based on contents
        # Code reference:
        # https://stackoverflow.com/questions/6337589/qlistwidget-adjust-size-to-content
        # ml: disabling this logic because I can't get it to work on HPD monitor.
        #     On my systems the values returned by the list object are also incorrect on
        #     creation, leading to a list object of size 15. count() always returns 0 until
        #     the list is actually displayed. The same is true for sizeHintForRow(0), which
        #     returns -1 until the widget is rendered. The widget claims to have a size of
        #     (100, 30), once it becomes visible the size is (535, 192).
        #     Leaving the framework here in case somebody figures out how to set this up
        #     properly.
        qList = self.form.baseList
        row = (qList.count() + qList.frameWidth()) * 15
        # qList.setMinimumHeight(row)
        Path.Log.debug(
            "baseList({}, {}) {} * {}".format(
                qList.size(), row, qList.count(), qList.sizeHintForRow(0)
            )
        )


class TaskPanelBaseLocationPage(TaskPanelPage):
    """Page controller for base locations. Uses PathGetPoint."""

    DataLocation = QtCore.Qt.ItemDataRole.UserRole

    def __init__(self, obj, features):
        super(TaskPanelBaseLocationPage, self).__init__(obj, features)

        # members initialized later
        self.editRow = None
        self.panelTitle = "Base Location"

    def getForm(self):
        self.formLoc = FreeCADGui.PySideUic.loadUi(":/panels/PageBaseLocationEdit.ui")
        if QtCore.qVersion()[0] == "4":
            self.formLoc.baseList.horizontalHeader().setResizeMode(QtGui.QHeaderView.Stretch)
        else:
            self.formLoc.baseList.horizontalHeader().setSectionResizeMode(QtGui.QHeaderView.Stretch)
        self.getPoint = PathGetPoint.TaskPanel(self.formLoc.addRemoveEdit)
        return self.formLoc

    def modifyStandardButtons(self, buttonBox):
        self.getPoint.buttonBox = buttonBox

    def getTitle(self, obj):
        return translate("PathOp", "Base Location")

    def getFields(self, obj):
        pass

    def setFields(self, obj):
        self.formLoc.baseList.blockSignals(True)
        self.formLoc.baseList.clearContents()
        self.formLoc.baseList.setRowCount(0)
        for location in self.obj.Locations:
            self.formLoc.baseList.insertRow(self.formLoc.baseList.rowCount())

            item = QtGui.QTableWidgetItem("%.2f" % location.x)
            item.setData(self.DataLocation, location.x)
            self.formLoc.baseList.setItem(self.formLoc.baseList.rowCount() - 1, 0, item)

            item = QtGui.QTableWidgetItem("%.2f" % location.y)
            item.setData(self.DataLocation, location.y)
            self.formLoc.baseList.setItem(self.formLoc.baseList.rowCount() - 1, 1, item)
        self.formLoc.baseList.resizeColumnToContents(0)
        self.formLoc.baseList.blockSignals(False)
        self.itemActivated()

    def removeLocation(self):
        deletedRows = []
        selected = self.formLoc.baseList.selectedItems()
        for item in selected:
            row = self.formLoc.baseList.row(item)
            if row not in deletedRows:
                deletedRows.append(row)
                self.formLoc.baseList.removeRow(row)
        self.updateLocations()
        FreeCAD.ActiveDocument.recompute()

    def updateLocations(self):
        Path.Log.track()
        locations = []
        for i in range(self.formLoc.baseList.rowCount()):
            x = self.formLoc.baseList.item(i, 0).data(self.DataLocation)
            y = self.formLoc.baseList.item(i, 1).data(self.DataLocation)
            location = FreeCAD.Vector(x, y, 0)
            locations.append(location)
        self.obj.Locations = locations

    def addLocation(self):
        self.getPoint.getPoint(self.addLocationAt)

    def addLocationAt(self, point, obj):
        if point:
            locations = self.obj.Locations
            locations.append(point)
            self.obj.Locations = locations
            FreeCAD.ActiveDocument.recompute()

    def editLocation(self):
        selected = self.formLoc.baseList.selectedItems()
        if selected:
            row = self.formLoc.baseList.row(selected[0])
            self.editRow = row
            x = self.formLoc.baseList.item(row, 0).data(self.DataLocation)
            y = self.formLoc.baseList.item(row, 1).data(self.DataLocation)
            start = FreeCAD.Vector(x, y, 0)
            self.getPoint.getPoint(self.editLocationAt, start)

    def editLocationAt(self, point, obj):
        if point:
            self.formLoc.baseList.item(self.editRow, 0).setData(self.DataLocation, point.x)
            self.formLoc.baseList.item(self.editRow, 1).setData(self.DataLocation, point.y)
            self.updateLocations()
            FreeCAD.ActiveDocument.recompute()

    def itemActivated(self):
        if self.formLoc.baseList.selectedItems():
            self.form.removeLocation.setEnabled(True)
            self.form.editLocation.setEnabled(True)
        else:
            self.form.removeLocation.setEnabled(False)
            self.form.editLocation.setEnabled(False)

    def registerSignalHandlers(self, obj):
        self.form.baseList.itemSelectionChanged.connect(self.itemActivated)
        self.formLoc.addLocation.clicked.connect(self.addLocation)
        self.formLoc.removeLocation.clicked.connect(self.removeLocation)
        self.formLoc.editLocation.clicked.connect(self.editLocation)

    def pageUpdateData(self, obj, prop):
        if prop in ["Locations"]:
            self.setFields(obj)


class TaskPanelHeightsPage(TaskPanelPage):
    """Page controller for heights."""

    def __init__(self, obj, features):
        super(TaskPanelHeightsPage, self).__init__(obj, features)

        # members initialized later
        self.clearanceHeight = None
        self.safeHeight = None
        self.panelTitle = "Heights"
        self.OpIcon = ":/icons/CAM_Heights.svg"
        self.setIcon(self.OpIcon)

    def getForm(self):
        return FreeCADGui.PySideUic.loadUi(":/panels/PageHeightsEdit.ui")

    def initPage(self, obj):
        self.safeHeight = PathGuiUtil.QuantitySpinBox(self.form.safeHeight, obj, "SafeHeight")
        self.clearanceHeight = PathGuiUtil.QuantitySpinBox(
            self.form.clearanceHeight, obj, "ClearanceHeight"
        )

    def getTitle(self, obj):
        return translate("PathOp", "Heights")

    def getFields(self, obj):
        self.safeHeight.updateProperty()
        self.clearanceHeight.updateProperty()

    def setFields(self, obj):
        self.safeHeight.updateWidget()
        self.clearanceHeight.updateWidget()

    def getSignalsForUpdate(self, obj):
        signals = []
        signals.append(self.form.safeHeight.editingFinished)
        signals.append(self.form.clearanceHeight.editingFinished)
        return signals

    def pageUpdateData(self, obj, prop):
        if prop in ["SafeHeight", "ClearanceHeight"]:
            self.setFields(obj)


class TaskPanelDepthsPage(TaskPanelPage):
    """Page controller for depths."""

    def __init__(self, obj, features):
        super(TaskPanelDepthsPage, self).__init__(obj, features)

        # members initialized later
        self.startDepth = None
        self.finalDepth = None
        self.finishDepth = None
        self.stepDown = None
        self.panelTitle = "Depths"
        self.OpIcon = ":/icons/CAM_Depths.svg"
        self.setIcon(self.OpIcon)

    def getForm(self):
        return FreeCADGui.PySideUic.loadUi(":/panels/PageDepthsEdit.ui")

    def haveStartDepth(self):
        return PathOp.FeatureDepths & self.features

    def haveFinalDepth(self):
        return (
            PathOp.FeatureDepths & self.features and not PathOp.FeatureNoFinalDepth & self.features
        )

    def haveFinishDepth(self):
        return PathOp.FeatureDepths & self.features and PathOp.FeatureFinishDepth & self.features

    def haveStepDown(self):
        return PathOp.FeatureStepDown & self.features

    def initPage(self, obj):

        if self.haveStartDepth():
            self.startDepth = PathGuiUtil.QuantitySpinBox(self.form.startDepth, obj, "StartDepth")
        else:
            self.form.startDepth.hide()
            self.form.startDepthLabel.hide()
            self.form.startDepthSet.hide()

        if self.haveFinalDepth():
            self.finalDepth = PathGuiUtil.QuantitySpinBox(self.form.finalDepth, obj, "FinalDepth")
        else:
            if self.haveStartDepth():
                self.form.finalDepth.setEnabled(False)
                self.form.finalDepth.setToolTip(
                    translate(
                        "PathOp",
                        "FinalDepth cannot be modified for this operation.\nIf it is necessary to set the FinalDepth manually please select a different operation.",
                    )
                )
            else:
                self.form.finalDepth.hide()
                self.form.finalDepthLabel.hide()
            self.form.finalDepthSet.hide()

        if self.haveStepDown():
            self.stepDown = PathGuiUtil.QuantitySpinBox(self.form.stepDown, obj, "StepDown")
        else:
            self.form.stepDown.hide()
            self.form.stepDownLabel.hide()

        if self.haveFinishDepth():
            self.finishDepth = PathGuiUtil.QuantitySpinBox(
                self.form.finishDepth, obj, "FinishDepth"
            )
        else:
            self.form.finishDepth.hide()
            self.form.finishDepthLabel.hide()

    def getTitle(self, obj):
        return translate("PathOp", "Depths")

    def getFields(self, obj):
        if self.haveStartDepth():
            self.startDepth.updateProperty()
        if self.haveFinalDepth():
            self.finalDepth.updateProperty()
        if self.haveStepDown():
            self.stepDown.updateProperty()
        if self.haveFinishDepth():
            self.finishDepth.updateProperty()

    def setFields(self, obj):
        if self.haveStartDepth():
            self.startDepth.updateWidget()
        if self.haveFinalDepth():
            self.finalDepth.updateWidget()
        if self.haveStepDown():
            self.stepDown.updateWidget()
        if self.haveFinishDepth():
            self.finishDepth.updateWidget()
        self.updateSelection(obj, FreeCADGui.Selection.getSelectionEx())

    def getSignalsForUpdate(self, obj):
        signals = []
        if self.haveStartDepth():
            signals.append(self.form.startDepth.editingFinished)
        if self.haveFinalDepth():
            signals.append(self.form.finalDepth.editingFinished)
        if self.haveStepDown():
            signals.append(self.form.stepDown.editingFinished)
        if self.haveFinishDepth():
            signals.append(self.form.finishDepth.editingFinished)
        return signals

    def registerSignalHandlers(self, obj):
        if self.haveStartDepth():
            self.form.startDepthSet.clicked.connect(
                lambda: self.depthSet(obj, self.startDepth, "StartDepth")
            )
        if self.haveFinalDepth():
            self.form.finalDepthSet.clicked.connect(
                lambda: self.depthSet(obj, self.finalDepth, "FinalDepth")
            )

    def pageUpdateData(self, obj, prop):
        if prop in ["StartDepth", "FinalDepth", "StepDown", "FinishDepth"]:
            self.setFields(obj)

    def depthSet(self, obj, spinbox, prop):
        z = self.selectionZLevel(FreeCADGui.Selection.getSelectionEx())
        if z is not None:
            Path.Log.debug("depthSet(%s, %s, %.2f)" % (obj.Label, prop, z))
            if spinbox.expression():
                obj.setExpression(prop, None)
                self.setDirty()
            spinbox.updateWidget(FreeCAD.Units.Quantity(z, FreeCAD.Units.Length))
            if spinbox.updateProperty():
                self.setDirty()
        else:
            Path.Log.info("depthSet(-)")

    def selectionZLevel(self, sel):
        if len(sel) == 1 and len(sel[0].SubObjects) == 1:
            sub = sel[0].SubObjects[0]
            if "Vertex" == sub.ShapeType:
                return sub.Z
            if Path.Geom.isHorizontal(sub):
                if "Edge" == sub.ShapeType:
                    return sub.Vertexes[0].Z
                if "Face" == sub.ShapeType:
                    return sub.BoundBox.ZMax
        return None

    def updateSelection(self, obj, sel):
        if self.selectionZLevel(sel) is not None:
            self.form.startDepthSet.setEnabled(True)
            self.form.finalDepthSet.setEnabled(True)
        else:
            self.form.startDepthSet.setEnabled(False)
            self.form.finalDepthSet.setEnabled(False)


class TaskPanelDiametersPage(TaskPanelPage):
    """Page controller for diameters."""

    def __init__(self, obj, features):
        super(TaskPanelDiametersPage, self).__init__(obj, features)

        # members initialized later
        self.clearanceHeight = None
        self.safeHeight = None

    def getForm(self):
        return FreeCADGui.PySideUic.loadUi(":/panels/PageDiametersEdit.ui")

    def initPage(self, obj):
        self.minDiameter = PathGuiUtil.QuantitySpinBox(self.form.minDiameter, obj, "MinDiameter")
        self.maxDiameter = PathGuiUtil.QuantitySpinBox(self.form.maxDiameter, obj, "MaxDiameter")

    def getTitle(self, obj):
        return translate("PathOp", "Diameters")

    def getFields(self, obj):
        self.minDiameter.updateProperty()
        self.maxDiameter.updateProperty()

    def setFields(self, obj):
        self.minDiameter.updateWidget()
        self.maxDiameter.updateWidget()

    def getSignalsForUpdate(self, obj):
        signals = []
        signals.append(self.form.minDiameter.editingFinished)
        signals.append(self.form.maxDiameter.editingFinished)
        return signals

    def pageUpdateData(self, obj, prop):
        if prop in ["MinDiameter", "MaxDiameter"]:
            self.setFields(obj)


class TaskPanel(object):
    """
    Generic TaskPanel implementation handling the standard Path operation layout.
    This class only implements the framework and takes care of bringing all pages up and down in a controller fashion.
    It implements the standard editor behaviour for OK, Cancel and Apply and tracks if the model is still in sync with
    the UI.
    However, all display and processing of fields is handled by the page controllers which are managed in a list. All
    event callbacks and framework actions are forwarded to the page controllers in turn and each page controller is
    expected to process all events concerning the data it manages.
    """

    def __init__(self, obj, deleteOnReject, opPage, selectionFactory):
        Path.Log.track(obj.Label, deleteOnReject, opPage, selectionFactory)
        FreeCAD.ActiveDocument.openTransaction(translate("PathOp", "AreaOp Operation"))
        self.obj = obj
        self.deleteOnReject = deleteOnReject
        self.featurePages = []
        self.parent = None

        # members initialized later
        self.clearanceHeight = None
        self.safeHeight = None
        self.startDepth = None
        self.finishDepth = None
        self.finalDepth = None
        self.stepDown = None
        self.buttonBox = None
        self.minDiameter = None
        self.maxDiameter = None

        features = obj.Proxy.opFeatures(obj)
        opPage.features = features

        if PathOp.FeatureBaseGeometry & features:
            if hasattr(opPage, "taskPanelBaseGeometryPage"):
                self.featurePages.append(opPage.taskPanelBaseGeometryPage(obj, features))
            else:
                self.featurePages.append(TaskPanelBaseGeometryPage(obj, features))

        if PathOp.FeatureLocations & features:
            if hasattr(opPage, "taskPanelBaseLocationPage"):
                self.featurePages.append(opPage.taskPanelBaseLocationPage(obj, features))
            else:
                self.featurePages.append(TaskPanelBaseLocationPage(obj, features))

        if PathOp.FeatureDepths & features or PathOp.FeatureStepDown & features:
            if hasattr(opPage, "taskPanelDepthsPage"):
                self.featurePages.append(opPage.taskPanelDepthsPage(obj, features))
            else:
                self.featurePages.append(TaskPanelDepthsPage(obj, features))

        if PathOp.FeatureHeights & features:
            if hasattr(opPage, "taskPanelHeightsPage"):
                self.featurePages.append(opPage.taskPanelHeightsPage(obj, features))
            else:
                self.featurePages.append(TaskPanelHeightsPage(obj, features))

        if PathOp.FeatureDiameters & features:
            if hasattr(opPage, "taskPanelDiametersPage"):
                self.featurePages.append(opPage.taskPanelDiametersPage(obj, features))
            else:
                self.featurePages.append(TaskPanelDiametersPage(obj, features))

        self.featurePages.append(opPage)

        for page in self.featurePages:
            page.parent = self  # save pointer to this current class as "parent"
            page.initPage(obj)
            page.onDirtyChanged(self.pageDirtyChanged)

        taskPanelLayout = Path.Preferences.defaultTaskPanelLayout()

        if taskPanelLayout < 2:
            opTitle = opPage.getTitle(obj)
            opPage.setTitle(translate("PathOp", "Operation"))
            toolbox = QtGui.QToolBox()
            if taskPanelLayout == 0:
                for page in self.featurePages:
                    toolbox.addItem(page.form, page.getTitle(obj))
                    itemIdx = toolbox.count() - 1
                    if page.icon:
                        toolbox.setItemIcon(itemIdx, QtGui.QIcon(page.icon))
                toolbox.setCurrentIndex(len(self.featurePages) - 1)
            else:
                for page in reversed(self.featurePages):
                    toolbox.addItem(page.form, page.getTitle(obj))
                    itemIdx = toolbox.count() - 1
                    if page.icon:
                        toolbox.setItemIcon(itemIdx, QtGui.QIcon(page.icon))
            toolbox.setWindowTitle(opTitle)
            if opPage.getIcon(obj):
                toolbox.setWindowIcon(QtGui.QIcon(opPage.getIcon(obj)))

            self.form = toolbox
        elif taskPanelLayout == 2:
            forms = []
            for page in self.featurePages:
                page.form.setWindowTitle(page.getTitle(obj))
                forms.append(page.form)
            self.form = forms
        elif taskPanelLayout == 3:
            forms = []
            for page in reversed(self.featurePages):
                page.form.setWindowTitle(page.getTitle(obj))
                forms.append(page.form)
            self.form = forms

        self.selectionFactory = selectionFactory
        self.obj = obj
        self.isdirty = deleteOnReject
        self.visibility = obj.ViewObject.Visibility
        obj.ViewObject.Visibility = True

    def isDirty(self):
        """isDirty() ... returns true if the model is not in sync with the UI anymore."""
        for page in self.featurePages:
            if page.isdirty:
                return True
        return self.isdirty

    def setClean(self):
        """setClean() ... set the receiver and all its pages clean."""
        self.isdirty = False
        for page in self.featurePages:
            page.setClean()

    def accept(self, resetEdit=True):
        """accept() ... callback invoked when user presses the task panel OK button."""
        self.preCleanup()
        if self.isDirty():
            self.panelGetFields()
        FreeCAD.ActiveDocument.commitTransaction()
        self.cleanup(resetEdit)

    def reject(self, resetEdit=True):
        """reject() ... callback invoked when user presses the task panel Cancel button."""
        self.preCleanup()
        FreeCAD.ActiveDocument.abortTransaction()
        if self.deleteOnReject:
            FreeCAD.ActiveDocument.openTransaction(translate("PathOp", "Uncreate AreaOp Operation"))
            try:
                PathUtil.clearExpressionEngine(self.obj)
                FreeCAD.ActiveDocument.removeObject(self.obj.Name)
            except Exception as ee:
                Path.Log.debug("{}\n".format(ee))
            FreeCAD.ActiveDocument.commitTransaction()
        self.cleanup(resetEdit)
        return True

    def preCleanup(self):
        for page in self.featurePages:
            page.onDirtyChanged(None)
        PathSelection.clear()
        FreeCADGui.Selection.removeObserver(self)
        self.obj.ViewObject.Proxy.clearTaskPanel()
        self.obj.ViewObject.Visibility = self.visibility

    def cleanup(self, resetEdit):
        """cleanup() ... implements common cleanup tasks."""
        self.panelCleanup()
        FreeCADGui.Control.closeDialog()
        if resetEdit:
            FreeCADGui.ActiveDocument.resetEdit()
        FreeCAD.ActiveDocument.recompute()

    def pageDirtyChanged(self, page):
        """pageDirtyChanged(page) ... internal callback"""
        self.buttonBox.button(QtGui.QDialogButtonBox.Apply).setEnabled(self.isDirty())

    def clicked(self, button):
        """clicked(button) ... callback invoked when the user presses any of the task panel buttons."""
        if button == QtGui.QDialogButtonBox.Apply:
            self.panelGetFields()
            self.setClean()
            FreeCAD.ActiveDocument.recompute()

    def modifyStandardButtons(self, buttonBox):
        """modifyStandarButtons(buttonBox) ... callback in case the task panel buttons need to be modified."""
        self.buttonBox = buttonBox
        for page in self.featurePages:
            page.modifyStandardButtons(buttonBox)
        self.pageDirtyChanged(None)

    def panelGetFields(self):
        """panelGetFields() ... invoked to trigger a complete transfer of UI data to the model."""
        Path.Log.track()
        for page in self.featurePages:
            page.pageGetFields()

    def panelSetFields(self):
        """panelSetFields() ... invoked to trigger a complete transfer of the model's properties to the UI."""
        Path.Log.track()
        self.obj.Proxy.sanitizeBase(self.obj)
        for page in self.featurePages:
            page.pageSetFields()

    def panelCleanup(self):
        """panelCleanup() ... invoked before the receiver is destroyed."""
        Path.Log.track()
        for page in self.featurePages:
            page.pageCleanup()

    def open(self):
        """open() ... callback invoked when the task panel is opened."""
        self.selectionFactory()
        FreeCADGui.Selection.addObserver(self)

    def getStandardButtons(self):
        """getStandardButtons() ... returns the Buttons for the task panel."""
        return (
            QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Apply | QtGui.QDialogButtonBox.Cancel
        )

    def setupUi(self):
        """setupUi() ... internal function to initialise all pages."""
        Path.Log.track(self.deleteOnReject)

        if self.deleteOnReject and PathOp.FeatureBaseGeometry & self.obj.Proxy.opFeatures(self.obj):
            sel = FreeCADGui.Selection.getSelectionEx()
            for page in self.featurePages:
                if getattr(page, "InitBase", True) and hasattr(page, "addBase"):
                    page.clearBase()
                    page.addBaseGeometry(sel)

        # Update properties based upon expressions in case expression value has changed
        for prp, expr in self.obj.ExpressionEngine:
            evalExpr = self.obj.evalExpression(expr)
            if not isinstance(evalExpr, (int, float, FreeCAD.Units.Quantity)):
                continue
            val = FreeCAD.Units.Quantity(evalExpr)
            value = val.Value if hasattr(val, "Value") else val
            prop = getattr(self.obj, prp)
            if hasattr(prop, "Value"):
                prop.Value = value
            else:
                prop = value

        self.panelSetFields()

        for page in self.featurePages:
            page.pageRegisterSignalHandlers()

    def updateData(self, obj, prop):
        """updateDate(obj, prop) ... callback invoked whenever a model's property is assigned a value."""
        # Path.Log.track(obj.Label, prop) # creates a lot of noise
        for page in self.featurePages:
            page.pageUpdateData(obj, prop)

    def needsFullSpace(self):
        return True

    def updateSelection(self):
        sel = FreeCADGui.Selection.getSelectionEx()
        for page in self.featurePages:
            page.updateSelection(self.obj, sel)

    # SelectionObserver interface
    def addSelection(self, doc, obj, sub, pnt):
        self.updateSelection()

    def removeSelection(self, doc, obj, sub):
        self.updateSelection()

    def setSelection(self, doc):
        self.updateSelection()

    def clearSelection(self, doc):
        self.updateSelection()


class CommandSetStartPoint:
    """Command to set the start point for an operation."""

    def GetResources(self):
        return {
            "Pixmap": "CAM_StartPoint",
            "MenuText": QT_TRANSLATE_NOOP("PathOp", "Start Point Selection"),
            "ToolTip": QT_TRANSLATE_NOOP("PathOp", "Selects the start point"),
        }

    def IsActive(self):
        if FreeCAD.ActiveDocument is None:
            return False
        sel = FreeCADGui.Selection.getSelection()
        if not sel:
            return False
        obj = sel[0]
        return obj and hasattr(obj, "StartPoint")

    def setpoint(self, point, o):
        FreeCADGui.Snapper.grid.off()
        obj = self.obj
        obj.StartPoint.x = point.x
        obj.StartPoint.y = point.y
        obj.StartPoint.z = obj.ClearanceHeight.Value
        obj.UseStartPoint = True
        obj.recompute()
        textPoint = f"{obj.StartPoint.x:.2f}, {obj.StartPoint.y:.2f}, {obj.StartPoint.z:.2f}"
        print(f"Set start point for operation {obj.Label} >>> {textPoint}")

    def Activated(self):
        self.obj = FreeCADGui.Selection.getSelection()[0]
        if not hasattr(FreeCADGui, "Snapper"):
            import DraftTools
        FreeCADGui.Snapper.getPoint(callback=self.setpoint)


def Create(res):
    """Create(res) ... generic implementation of a create function.
    res is an instance of CommandResources. It is not expected that the user invokes
    this function directly, but calls the Activated() function of the Command object
    that is created in each operations Gui implementation."""
    FreeCAD.ActiveDocument.openTransaction("Create %s" % res.name)
    if res.job is None:
        FreeCAD.ActiveDocument.abortTransaction()
        raise ValueError("No job selected. Operation creation aborted.")
    try:
        obj = res.objFactory(res.name, obj=None, parentJob=res.job)
        if obj.Proxy:
            obj.ViewObject.Proxy = ViewProvider(obj.ViewObject, res)
            obj.ViewObject.Visibility = True
            FreeCAD.ActiveDocument.commitTransaction()

            obj.ViewObject.Document.setEdit(obj.ViewObject, 0)
            return obj
    except PathUtils.PathNoTCExistsException:
        msg = translate("PathOp", "No suitable tool controller found.\nAborting op creation")
        diag = QtGui.QMessageBox(QtGui.QMessageBox.Warning, "Error", msg)
        diag.setWindowModality(QtCore.Qt.ApplicationModal)
        diag.exec_()
    except PathOp.PathNoTCException:
        Path.Log.warning(translate("PathOp", "No tool controller, aborting op creation"))

    FreeCAD.ActiveDocument.abortTransaction()
    FreeCAD.ActiveDocument.recompute()
    return None


class CommandPathOp:
    """Generic, data driven implementation of a Path operation creation command.
    Instances of this class are stored in all Path operation Gui modules and can
    be used to create said operations with view providers and all."""

    def __init__(self, resources):
        self.res = resources

    def GetResources(self):
        ress = {
            "Pixmap": self.res.pixmap,
            "MenuText": self.res.menuText,
            "ToolTip": self.res.toolTip,
        }
        if self.res.accelKey:
            ress["Accel"] = self.res.accelKey
        return ress

    def IsActive(self):
        if FreeCAD.ActiveDocument is not None:
            for o in FreeCAD.ActiveDocument.Objects:
                if o.Name[:3] == "Job":
                    return True
        return False

    def Activated(self):
        jobs = PathUtils.GetJobs()
        if not jobs:
            return
        job = PathUtils.UserInput.chooseJob(jobs)
        if job is None:
            return  # Abort if no job selected or canceled
        self.res.job = job
        return Create(self.res)

    def setJob(self, job):
        self.res.job = job
        return Create(self.res)


class CommandResources:
    """POD class to hold command specific resources."""

    def __init__(self, name, objFactory, opPageClass, pixmap, menuText, accelKey, toolTip):
        self.name = name
        self.objFactory = objFactory
        self.opPageClass = opPageClass
        self.pixmap = pixmap
        self.menuText = menuText
        self.accelKey = accelKey
        self.toolTip = toolTip
        self.job = None


def SetupOperation(name, objFactory, opPageClass, pixmap, menuText, toolTip, setupProperties=None):
    """SetupOperation(name, objFactory, opPageClass, pixmap, menuText, toolTip, setupProperties=None)
    Creates an instance of CommandPathOp with the given parameters and registers the command with FreeCAD.
    When activated it creates a model with proxy (by invoking objFactory), assigns a view provider to it
    (see ViewProvider in this module) and starts the editor specifically for this operation (driven by opPageClass).
    This is an internal function that is automatically called by the initialisation code for each operation.
    It is not expected to be called manually.
    """

    res = CommandResources(name, objFactory, opPageClass, pixmap, menuText, None, toolTip)

    command = CommandPathOp(res)
    FreeCADGui.addCommand("CAM_%s" % name.replace(" ", "_"), command)

    if setupProperties is not None:
        PathSetupSheet.RegisterOperation(name, objFactory, setupProperties)

    return command


FreeCADGui.addCommand("CAM_SetStartPoint", CommandSetStartPoint())

FreeCAD.Console.PrintLog("Loading PathOpGui… done\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# This package aggregates tool bit classes.

# Import the base class and all concrete shape classes
from .models.base import ToolBit
from .models.ballend import ToolBitBallend
from .models.bullnose import ToolBitBullnose
from .models.chamfer import ToolBitChamfer
from .models.custom import ToolBitCustom
from .models.dovetail import ToolBitDovetail
from .models.drill import ToolBitDrill
from .models.endmill import ToolBitEndmill
from .models.radius import ToolBitRadius
from .models.probe import ToolBitProbe
from .models.reamer import ToolBitReamer
from .models.slittingsaw import ToolBitSlittingSaw
from .models.taperedballnose import ToolBitTaperedBallNose
from .models.tap import ToolBitTap
from .models.threadmill import ToolBitThreadMill
from .models.vbit import ToolBitVBit

# Define __all__ for explicit public interface
__all__ = [
    "ToolBit",
    "ToolBitBallend",
    "ToolBitBullnose",
    "ToolBitChamfer",
    "ToolBitCustom",
    "ToolBitDovetail",
    "ToolBitDrill",
    "ToolBitEndmill",
    "ToolBitRadius",
    "ToolBitProbe",
    "ToolBitReamer",
    "ToolBitSlittingSaw",
    "ToolBitTaperedBallNose",
    "ToolBitTap",
    "ToolBitThreadMill",
    "ToolBitVBit",
]



# ***************************************************************************
# *   Copyright (c) 2015 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM element geometry 1D document object"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package element_geometry1D
#  \ingroup FEM
#  \brief element geometry 1D object

from FreeCAD import Base
from . import base_femelement
from . import base_fempythonobject

_PropHelper = base_fempythonobject._PropHelper


class ElementGeometry1D(base_femelement.BaseFemElement):
    """
    The ElementGeometry1D object
    """

    Type = "Fem::ElementGeometry1D"

    def __init__(self, obj):
        super().__init__(obj)

    def _get_properties(self):
        prop = super()._get_properties()

        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="RectWidth",
                group="RectBeamSection",
                doc="Set width of the rectangular beam elements",
                value=0.0,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="RectHeight",
                group="RectBeamSection",
                doc="Set height of there ctangular beam elements",
                value=0.0,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="CircDiameter",
                group="CircBeamSection",
                doc="Set diameter of the circular beam elements",
                value=0.0,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="PipeDiameter",
                group="PipeBeamSection",
                doc="Set outer diameter of the pipe beam elements",
                value=0.0,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="PipeThickness",
                group="PipeBeamSection",
                doc="Set thickness of the pipe beam elements",
                value=0.0,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="Axis1Length",
                group="EllipticalBeamSection",
                doc="Set first principal axis length of the elliptical beam elements",
                value=0.0,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="Axis2Length",
                group="EllipticalBeamSection",
                doc="Set second principal axis length of the elliptical beam elements",
                value=0.0,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="BoxWidth",
                group="BoxBeamSection",
                doc="Set width of the box beam elements",
                value=0.0,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="BoxHeight",
                group="BoxBeamSection",
                doc="Set height of the box beam elements",
                value=0.0,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="BoxT1",
                group="BoxBeamSection",
                doc="Set thickness parameter t1 of the box beam elements",
                value=0.0,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="BoxT2",
                group="BoxBeamSection",
                doc="Set thickness parameter t2 of the box beam elements",
                value=0.0,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="BoxT3",
                group="BoxBeamSection",
                doc="Set thickness parameter t3 of the box beam elements",
                value=0.0,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyLength",
                name="BoxT4",
                group="BoxBeamSection",
                doc="Set thickness parameter t4 of the box beam elements",
                value=0.0,
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyEnumeration",
                name="SectionType",
                group="BeamSection",
                doc="Select beam section type",
                value=["Rectangular", "Circular", "Pipe", "Elliptical", "Box"],
            )
        )
        prop.append(
            _PropHelper(
                type="App::PropertyArea",
                name="TrussArea",
                group="TrussSection",
                doc="Set cross-sectional area of truss elements\n"
                + "(used if bending stiffness is excluded in the solver)",
                value=10.0,
            )
        )
        return prop

    def onDocumentRestored(self, obj):
        # update old project with new properties
        super().onDocumentRestored(obj)

        for prop in self._get_properties():
            try:
                obj.getPropertyByName(prop.name)
            except Base.PropertyError:
                prop.add_to_object(obj)

            if prop.name == "SectionType":
                # refresh the list of known section types for old projects
                obj.SectionType = prop.value
